from os.path import join
from re import compile as re_compile
from typing import List

from supervisely.nn.artifacts.artifacts import BaseTrainArtifacts


class RITM(BaseTrainArtifacts):
    """Train/serve artifacts handler for RITM interactive segmentation models."""

    def __init__(self, team_id: int):
        """See :class:`~supervisely.nn.artifacts.artifacts.BaseTrainArtifacts` for params."""
        super().__init__(team_id)

        self._app_name = "Train RITM"
        self._slug = "supervisely-ecosystem/ritm-training/supervisely/train"
        self._serve_app_name = None
        self._serve_slug = None
        self._framework_name = "RITM"
        self._framework_folder = "/RITM_training"
        self._weights_folder = "checkpoints"
        self._task_type = None
        self._info_file = "info/ui_state.json"
        self._weights_ext = ".pth"
        self._pattern = re_compile(r"^/RITM_training/\d+_[^/]+/?$")
        self._available_task_types: List[str] = ["interactive segmentation"]
        self._require_runtime = False
        self._has_benchmark_evaluation = False

    def get_task_id(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        if len(parts) < 3:
            raise ValueError(f"Invalid path: {artifacts_folder}")
        session_id, _ = parts[2].split("_", 1)
        return session_id

    def get_project_name(self, artifacts_folder: str) -> str:
        parts = artifacts_folder.split("/")
        if len(parts) < 3:
            raise ValueError(f"Invalid path: {artifacts_folder}")
        _, project_name = parts[2].split("_", 1)
        return project_name

    def get_task_type(self, artifacts_folder: str) -> str:
        info_path = join(artifacts_folder, self._info_file)
        task_type = "undefined"
        for file_info in self._get_file_infos():
            if file_info.path == info_path:
                json_data = self._fetch_json_from_path(file_info.path)
                task_type = json_data.get("segmentationType", "undefined")
                if task_type is not None:
                    task_type = task_type.lower()
                break
        return task_type

    def get_weights_path(self, artifacts_folder: str) -> str:
        return join(artifacts_folder, self._weights_folder)

    def get_config_path(self, artifacts_folder: str) -> str:
        return None



from __future__ import annotations

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import (
    ChartWidget,
    MarkdownWidget,
    TableWidget,
)


class Speedtest(DetectionVisMetric):
    """Inference speed and FPS tables/charts for detection reports."""

    MARKDOWN_INTRO = "speedtest_intro"
    TABLE_MARKDOWN = "speedtest_table"
    TABLE = "speedtest"
    CHART_MARKDOWN = "speedtest_chart"
    CHART = "speedtest"

    def is_empty(self) -> bool:
        return not self.eval_result.speedtest_info

    @property
    def num_batche_sizes(self) -> int:
        if self.is_empty():
            return 0
        return len(self.eval_result.speedtest_info.get("speedtest", []))

    @property
    def intro_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_speedtest_intro
        text = text.format(
            self.eval_result.speedtest_info["model_info"]["device"],
            self.eval_result.speedtest_info["model_info"]["hardware"],
            self.eval_result.speedtest_info["model_info"]["runtime"],
        )
        return MarkdownWidget(name=self.MARKDOWN_INTRO, title="Inference Speed", text=text)

    @property
    def table_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_speedtest_table
        text = text.format(self.eval_result.speedtest_info["speedtest"][0]["num_iterations"])
        return MarkdownWidget(name=self.TABLE_MARKDOWN, title="Speedtest Table", text=text)

    @property
    def table(self) -> TableWidget:
        columns = [" ", "Inference time", "FPS"]
        content = []
        temp_res = {}
        max_fps = 0
        for test in self.eval_result.speedtest_info["speedtest"]:
            batch_size = test["batch_size"]

            ms = round(test["benchmark"]["total"], 2)
            fps = round(1000 / test["benchmark"]["total"] * batch_size)
            row = [f"Batch size {batch_size}", ms, fps]
            temp_res[batch_size] = row
            max_fps = max(max_fps, fps)

        # sort by batch size
        temp_res = dict(sorted(temp_res.items()))
        for row in temp_res.values():
            content.append({"row": row, "id": row[0], "items": row})

        columns_options = [
            {"disableSort": True},  # , "ustomCell": True},
            {"subtitle": "ms", "tooltip": "Milliseconds for batch images", "postfix": "ms"},
            {
                "subtitle": "imgs/sec",
                "tooltip": "Frames (images) per second",
                "postfix": "fps",
                "maxValue": max_fps,
            },
        ]
        data = {"columns": columns, "columnsOptions": columns_options, "content": content}
        table = TableWidget(name=self.TABLE, data=data)
        table.main_column = "Batch size"
        table.fixed_columns = 1
        table.show_header_controls = False
        return table

    @property
    def chart_md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_speedtest_chart
        return MarkdownWidget(name=self.CHART_MARKDOWN, title="Speedtest Chart", text=text)

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(name=self.CHART, figure=self._get_figure())

    def _get_figure(self):  #  -> go.Figure:
        import plotly.graph_objects as go  # pylint: disable=import-error
        from plotly.subplots import make_subplots  # pylint: disable=import-error

        fig = make_subplots(cols=2)

        ms_color = "#e377c2"
        fps_color = "#17becf"

        temp_res = {}
        for test in self.eval_result.speedtest_info["speedtest"]:
            batch_size = test["batch_size"]

            std = test["benchmark_std"]["total"]
            ms = test["benchmark"]["total"]
            fps = round(1000 / test["benchmark"]["total"] * batch_size)

            ms_line = temp_res.setdefault("ms", {})
            fps_line = temp_res.setdefault("fps", {})
            ms_std_line = temp_res.setdefault("ms_std", {})

            ms_line[batch_size] = ms
            fps_line[batch_size] = fps
            ms_std_line[batch_size] = round(std, 2)

        fig.add_trace(
            go.Scatter(
                x=list(temp_res["ms"].keys()),
                y=list(temp_res["ms"].values()),
                name="Inference time (ms)",
                line=dict(color=ms_color),
                customdata=list(temp_res["ms_std"].values()),
                error_y=dict(
                    type="data",
                    array=list(temp_res["ms_std"].values()),
                    visible=True,
                    color="rgba(227, 119, 194, 0.7)",
                ),
                hovertemplate="Batch Size: %{x}<br>Time: %{y:.2f} ms<br> Standard deviation: %{customdata:.2f} ms<extra></extra>",
            ),
            col=1,
            row=1,
        )
        fig.add_trace(
            go.Scatter(
                x=list(temp_res["fps"].keys()),
                y=list(temp_res["fps"].values()),
                name="FPS",
                line=dict(color=fps_color),
                hovertemplate="Batch Size: %{x}<br>FPS: %{y:.2f}<extra></extra>",
            ),
            col=2,
            row=1,
        )

        fig.update_xaxes(title_text="Batch size", col=1, dtick=1)
        fig.update_xaxes(title_text="Batch size", col=2, dtick=1)

        fig.update_yaxes(title_text="Time (ms)", col=1)
        fig.update_yaxes(title_text="FPS", col=2)
        fig.update_layout(height=400)

        return fig



# coding: utf-8


class SingleImageInferenceInterface:
    """Abstract interface for single-image inference; subclasses implement inference() and get_out_meta()."""

    def inference(self, image, ann):
        raise NotImplementedError()

    def get_out_meta(self):
        raise NotImplementedError()



from supervisely.app.widgets import (
    Card,
    Text,
    Select,
    Field,
    Container,
    ObjectClassView,
    Checkbox,
    ObjectClassesList,
    Widget,
)
from supervisely.annotation.obj_class import ObjClass


from supervisely.geometry.bitmap import Bitmap
from supervisely.geometry.cuboid import Cuboid
from supervisely.geometry.point import Point
from supervisely.geometry.polygon import Polygon
from supervisely.geometry.polyline import Polyline
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.graph import GraphNodes
from supervisely.geometry.any_geometry import AnyGeometry
from supervisely.geometry.cuboid_3d import Cuboid3d
from supervisely.geometry.pointcloud import Pointcloud
from supervisely.geometry.point_3d import Point3d
from supervisely.geometry.multichannel_bitmap import MultichannelBitmap
from supervisely.geometry.closed_surface_mesh import ClosedSurfaceMesh

# model_classes = Container(
#     [
#         Checkbox(ObjectClassView(ObjClass("person", AnyGeometry))),
#         Checkbox(ObjectClassView(ObjClass("person", Rectangle))),
#         Checkbox(ObjectClassView(ObjClass("person", Polygon))),
#         Checkbox(ObjectClassView(ObjClass("person", Bitmap))),
#         Checkbox(ObjectClassView(ObjClass("person", Point))),
#         Checkbox(ObjectClassView(ObjClass("person", Polyline))),
#         Checkbox(ObjectClassView(ObjClass("person", Cuboid))),
#         # ObjClassView(ObjClass("person", GraphNodes, geometry_config={"1": "2"})),
#         Checkbox(ObjectClassView(ObjClass("person", Point3d))),
#         Checkbox(ObjectClassView(ObjClass("person", Cuboid3d))),
#         Checkbox(ObjectClassView(ObjClass("person", Pointcloud))),
#         Checkbox(ObjectClassView(ObjClass("person", Cuboid3d))),
#         Checkbox(ObjectClassView(ObjClass("person", Pointcloud))),
#         Checkbox(ObjectClassView(ObjClass("person", ClosedSurfaceMesh))),
#         Checkbox(ObjectClassView(ObjClass("person", MultichannelBitmap))),
#     ],
#     direction="horizontal",
#     overflow="wrap",
# )


_classes = [
    ObjClass("person1", AnyGeometry),
    ObjClass("person2", Rectangle),
    ObjClass("person3", Polygon),
    ObjClass("person4", Bitmap),
    ObjClass("person5", Point),
    ObjClass("person6", Polyline),
    ObjClass("person7", Cuboid),
    ObjClass("person8", Point3d),
    ObjClass("person9", Cuboid3d),
    ObjClass("person10", Pointcloud),
    ObjClass("person11", Cuboid3d),
    ObjClass("person12", Pointcloud),
    ObjClass("person13", ClosedSurfaceMesh),
    ObjClass("person14", MultichannelBitmap),
]

model_classes = ObjectClassesList(object_classes=_classes, columns=3, selectable=True)

text = Text("123")
classes_card = Card(
    "Model classes",
    "Model predicts the following classes",
    content=model_classes,
)

classes_layout = Container(
    [
        classes_card,
    ],
    direction="vertical",
)



# coding: utf-8
from __future__ import annotations

import json
import uuid
from copy import deepcopy
from typing import Dict, List, Optional, Tuple
from uuid import UUID

from supervisely._utils import take_with_default
from supervisely.project.project_meta import ProjectMeta
from supervisely.video_annotation.constants import (
    DESCRIPTION,
    FRAMES,
    FRAMES_COUNT,
    IMG_SIZE,
    IMG_SIZE_HEIGHT,
    IMG_SIZE_WIDTH,
    KEY,
    OBJECTS,
    TAGS,
    VIDEO_ID,
)
from supervisely.video_annotation.frame_collection import FrameCollection
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_figure import VideoFigure
from supervisely.video_annotation.video_object_collection import VideoObjectCollection
from supervisely.video_annotation.video_tag_collection import VideoTagCollection


class VideoAnnotation:
    """
    Annotation for a single video in Supervisely format.

    Stores video-level tags, objects and per-frame figures, and supports JSON (de)serialization
    via :meth:`to_json` / :meth:`from_json`.
    """

    def __init__(
        self,
        img_size: Tuple[int, int],
        frames_count: int,
        objects: Optional[VideoObjectCollection] = None,
        frames: Optional[FrameCollection] = None,
        tags: Optional[VideoTagCollection] = None,
        description: Optional[str] = "",
        key: Optional[UUID] = None,
    ):
        """
        VideoAnnotation for a single video. :class:`~supervisely.video_annotation.video_annotation.VideoAnnotation` object is immutable.

        :param img_size: Size of the image (height, width).
        :type img_size: Tuple[int, int] or List[int, int]
        :param frames_count: Number of frames in VideoAnnotation.
        :type frames_count: int
        :param objects: VideoObjectCollection object.
        :type objects: :class:`~supervisely.video_annotation.video_object_collection.VideoObjectCollection`, optional
        :param frames: FrameCollection object.
        :type frames: :class:`~supervisely.video_annotation.frame_collection.FrameCollection`, optional
        :param tags: VideoTagCollection object.
        :type tags: :class:`~supervisely.video_annotation.video_tag_collection.VideoTagCollection`, optional
        :param description: Video description.
        :type description: str, optional
        :param key: UUID object.
        :type key: UUID, optional
        :raises TypeError: if img_size is not tuple or list

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.video_annotation.video_tag import VideoTag
                from supervisely.video_annotation.video_tag_collection import VideoTagCollection

                # Simple VideoAnnotation example
                height, width = 500, 700
                frames_count = 10
                video_ann = sly.VideoAnnotation((height, width), frames_count)
                print(video_ann.to_json())
                # Output: {
                #     "size": {
                #         "height": 500,
                #         "width": 700
                #     },
                #     "description": "",
                #     "key": "abef780b01ad4063b4b961ab2ba2f410",
                #     "tags": [],
                #     "objects": [],
                #     "frames": [],
                #     "framesCount": 10
                # }

                # More complex VideoAnnotation example
                height, width = 500, 700
                frames_count = 1
                # VideoObjectCollection
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                objects = sly.VideoObjectCollection([video_obj_car])

                # FrameCollection
                fr_index = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
                frame = sly.Frame(fr_index, figures=[video_figure_car])
                frames = sly.FrameCollection([frame])

                # VideoTagCollection
                meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
                vid_tag = VideoTag(meta_car, value='acura')
                video_tags = VideoTagCollection([vid_tag])

                # Description
                descr = 'car example'

                video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames, video_tags, descr)
                print(video_ann.to_json())
                # Output: {
                #     "size": {
                #         "height": 500,
                #         "width": 700
                #     },
                #     "description": "car example",
                #     "key": "a85b282e5e174e7ebad6f878b6919244",
                #     "tags": [
                #         {
                #             "name": "car_tag",
                #             "value": "acura",
                #             "key": "540a8212b0344788953996cea220ea8b"
                #         }
                #     ],
                #     "objects": [
                #         {
                #             "key": "7c74b8a495044ea0ac127f32751c8f5c",
                #             "classTitle": "car",
                #             "tags": []
                #         }
                #     ],
                #     "frames": [
                #         {
                #             "index": 7,
                #             "figures": [
                #                 {
                #                     "key": "82dcbf2e3c5f42a99eeea2ad34173793",
                #                     "objectKey": "7c74b8a495044ea0ac127f32751c8f5c",
                #                     "geometryType": "rectangle",
                #                     "geometry": {
                #                         "points": {
                #                             "exterior": [
                #                                 [
                #                                     0,
                #                                     0
                #                                 ],
                #                                 [
                #                                     100,
                #                                     100
                #                                 ]
                #                             ],
                #                             "interior": []
                #                         }
                #                     }
                #                 }
                #             ]
                #         }
                #     ],
                #     "framesCount": 1
                # }
        """
        if not isinstance(img_size, (tuple, list)):
            raise TypeError(
                '{!r} has to be a tuple or a list. Given type "{}".'.format(
                    "img_size", type(img_size)
                )
            )
        self._img_size = tuple(img_size)
        self._frames_count = frames_count

        self._description = description
        self._tags = take_with_default(tags, VideoTagCollection())
        self._objects = take_with_default(objects, VideoObjectCollection())
        self._frames = take_with_default(frames, FrameCollection())
        self._key = take_with_default(key, uuid.uuid4())

        self.validate_figures_bounds()

    @property
    def img_size(self) -> Tuple[int, int]:
        """
        Size of the image (height, width).

        :returns: Image size
        :rtype: :class:`Tuple[int, int]`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 1
                video_ann = sly.VideoAnnotation((height, width), frames_count)
                print(video_ann.img_size)
                # Output: (500, 700)
        """
        return deepcopy(self._img_size)

    @property
    def frames_count(self) -> int:
        """
        Number of frames.

        :returns: Frames count
        :rtype: int

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 15
                video_ann = sly.VideoAnnotation((height, width), frames_count)
                print(video_ann.frames_count)
                # Output: 15
        """
        return self._frames_count

    @property
    def objects(self) -> VideoObjectCollection:
        """
        VideoAnnotation objects.

        :returns: VideoObjectCollection object.
        :rtype: VideoObjectCollection

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 1
                # VideoObjectCollection
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                objects = sly.VideoObjectCollection([video_obj_car])
                video_ann = sly.VideoAnnotation((height, width), frames_count, objects)
                print(video_ann.objects.to_json())
                # Output: [
                #     {
                #         "key": "79fc07a4a6ca4b2796279bc033b9ec9a",
                #         "classTitle": "car",
                #         "tags": []
                #     }
                # ]
        """
        return self._objects

    @property
    def frames(self) -> FrameCollection:
        """
        VideoAnnotation frames.

        :returns: FrameCollection object.
        :rtype: FrameCollection

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 1
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                objects = sly.VideoObjectCollection([video_obj_car])
                fr_index = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
                frame = sly.Frame(fr_index, figures=[video_figure_car])
                frames = sly.FrameCollection([frame])

                video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames)
                print(video_ann.frames.to_json())
                # Output: [
                #     {
                #         "index": 7,
                #         "figures": [
                #             {
                #                 "key": "2842f561b1924f6abd6ab6f696ed9b65",
                #                 "objectKey": "7f30fa9b78444ad69e02b37edbf9a902",
                #                 "geometryType": "rectangle",
                #                 "geometry": {
                #                     "points": {
                #                         "exterior": [
                #                             [
                #                                 0,
                #                                 0
                #                             ],
                #                             [
                #                                 100,
                #                                 100
                #                             ]
                #                         ],
                #                         "interior": []
                #                     }
                #                 }
                #             }
                #         ]
                #     }
                # ]
        """
        return self._frames

    @property
    def figures(self) -> List[VideoFigure]:
        """
        VideoAnnotation figures.

        :returns: List of VideoFigures from all frames in VideoAnnotation.
        :rtype: List[:class:`~supervisely.video_annotation.video_figure.VideoFigure`]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 1
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                objects = sly.VideoObjectCollection([video_obj_car])
                fr_index = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
                frame = sly.Frame(fr_index, figures=[video_figure_car])
                frames = sly.FrameCollection([frame])

                video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames)
                print(len(video_ann.figures)) # 1
        """
        return self.frames.figures

    @property
    def tags(self) -> VideoTagCollection:
        """
        VideoAnnotation tags.

        :returns: VideoTagCollection object.
        :rtype: :class:`~supervisely.video_annotation.video_tag_collection.VideoTagCollection`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 1
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                objects = sly.VideoObjectCollection([video_obj_car])
                fr_index = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
                frame = sly.Frame(fr_index, figures=[video_figure_car])
                frames = sly.FrameCollection([frame])
                meta_car = sly.TagMeta('car_tag', sly.TagValueType.ANY_STRING)
                from supervisely.video_annotation.video_tag import VideoTag
                vid_tag = VideoTag(meta_car, value='acura')
                from supervisely.video_annotation.video_tag_collection import VideoTagCollection
                tags = VideoTagCollection([vid_tag])

                video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames, tags)
                print(video_ann.tags.to_json())
                # Output: [
                #     {
                #         "name": "car_tag",
                #         "value": "acura",
                #         "key": "c63e8259589a4fa5b4fb15a48c1f6a63"
                #     }
                # ]
        """
        return self._tags

    def key(self) -> UUID:
        """
        Annotation key value.

        :returns: Key value of annotation object.
        :rtype: UUID

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 1
                # VideoObjectCollection
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                objects = sly.VideoObjectCollection([video_obj_car])
                video_ann = sly.VideoAnnotation((height, width), frames_count, objects)

                print(video_ann.key())
                # Output: 6e5bd622-4d7b-45ee-8bc5-807d5a5e2134
        """

        return self._key

    @property
    def description(self) -> str:
        """
        Video description.

        :returns: Video description
        :rtype: str

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 1
                descr = 'example'
                video_ann = sly.VideoAnnotation((height, width), frames_count, description=descr)
                print(video_ann.description) # example
        """
        return self._description

    # def get_tags_on_frame(self, frame_index: int) -> VideoTagCollection:
    #     """
    #     Get all existing video tags from frame of video.

    #     :param frame_index: Video frame index.
    #     :type frame_index: int
    #     :returns: Tags from the given frame.
    #     :rtype: :class:`~supervisely.video_annotation.video_tag_collection.VideoTagCollection`

    #     :Usage Example:

    #      .. code-block:: python

    #         import supervisely as sly

    #         height, width = 50, 700
    #         frames_count = 1
    #         obj_class_car = sly.ObjClass('car', sly.Rectangle)
    #         video_obj_car = sly.VideoObject(obj_class_car)
    #         objects = sly.VideoObjectCollection([video_obj_car])
    #         fr_index = 7
    #         geometry = sly.Rectangle(10, 10, 40, 40)
    #         video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
    #         frame = sly.Frame(fr_index, figures=[video_figure_car])
    #         frames = sly.FrameCollection([frame])
    #         meta_car = sly.TagMeta('car', sly.TagValueType.NONE)
    #         tag_car = sly.VideoTag(meta_car, frame_range=(fr_index, fr_index))
    #         tags = sly.VideoTagCollection([tag_car])

    #         video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames, tags)
    #         tags_on_frame = video_ann.get_tags_on_frame(fr_index)
    #         print(len(tags_on_frame))
    #         # Output: 1
    #     """
    #     tags = []
    #     for tag in self._tags:
    #         if frame_index >= tag.frame_range[0] and frame_index <= tag.frame_range[1]:
    #             tags.append(tag)
    #     return VideoTagCollection(tags)

    # def get_objects_on_frame(self, frame_index: int) -> VideoObjectCollection:
    #     """
    #     Get all existing video objects from frame of video.

    #     :param frame_index: Video frame index.
    #     :type frame_index: int
    #     :returns: Objects from the given frame.
    #     :rtype: :class:`~supervisely.video_annotation.video_object_collection.VideoObjectCollection`

    #     :Usage Example:

    #      .. code-block:: python

    #         import supervisely as sly

    #         height, width = 50, 700
    #         frames_count = 1
    #         obj_class_car = sly.ObjClass('car', sly.Rectangle)
    #         video_obj_car = sly.VideoObject(obj_class_car)
    #         objects = sly.VideoObjectCollection([video_obj_car])
    #         fr_index = 7
    #         geometry = sly.Rectangle(10, 10, 40, 40)
    #         video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
    #         frame = sly.Frame(fr_index, figures=[video_figure_car])
    #         frames = sly.FrameCollection([frame])

    #         video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames)
    #         objs_on_frame = video_ann.get_objects_on_frame(fr_index)
    #         print(len(objs_on_frame))
    #         # Output: 1
    #     """
    #     frame = self._frames.get(frame_index, None)
    #     if frame is None:
    #         raise ValueError(f"No frame with index {frame_index} in annotation.")
    #     frame_objects = {}
    #     for fig in frame.figures:
    #         if fig.parent_object.key() not in frame_objects.keys():
    #             frame_objects[fig.parent_object.key()] = fig.parent_object
    #     return VideoObjectCollection(list(frame_objects.values()))

    def validate_figures_bounds(self) -> None:
        """
        Checks if image contains figures from all frames in collection.

        :raises :class:`~supervisely.video_annotation.video_figure.OutOfImageBoundsException`: if figure is out of image bounds
        :returns: None
        :rtype: None

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 50, 700
                frames_count = 1
                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                objects = sly.VideoObjectCollection([video_obj_car])
                fr_index = 7
                geometry = sly.Rectangle(0, 0, 100, 100)
                video_figure_car = sly.VideoFigure(video_obj_car, geometry, fr_index)
                frame = sly.Frame(fr_index, figures=[video_figure_car])
                frames = sly.FrameCollection([frame])

                video_ann = sly.VideoAnnotation((height, width), frames_count, objects, frames)
                video_ann.validate_figures_bounds()
                # raise OutOfImageBoundsException("Figure is out of image bounds")
        """
        for frame in self.frames:
            frame.validate_figures_bounds(self.img_size)

    def to_json(self, key_id_map: Optional[KeyIdMap] = None) -> Dict:
        """
        Convert the VideoAnnotation to a json dict. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: Json format as a dict
        :rtype: Dict

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 10
                video_ann = sly.VideoAnnotation((height, width), frames_count)
                print(video_ann.to_json())
                # Output: {
                #     "size": {
                #         "height": 500,
                #         "width": 700
                #     },
                #     "description": "",
                #     "key": "abef780b01ad4063b4b961ab2ba2f410",
                #     "tags": [],
                #     "objects": [],
                #     "frames": [],
                #     "framesCount": 10
                # }
        """
        res_json = {
            IMG_SIZE: {
                IMG_SIZE_HEIGHT: int(self.img_size[0]),
                IMG_SIZE_WIDTH: int(self.img_size[1]),
            },
            DESCRIPTION: self.description,
            KEY: self.key().hex,
            TAGS: self.tags.to_json(key_id_map),
            OBJECTS: self.objects.to_json(key_id_map),
            FRAMES: self.frames.to_json(key_id_map),
            FRAMES_COUNT: self.frames_count,
        }

        if key_id_map is not None:
            video_id = key_id_map.get_video_id(self.key())
            if video_id is not None:
                res_json[VIDEO_ID] = video_id

        return res_json

    @classmethod
    def from_json(
        cls,
        data: Dict,
        project_meta: ProjectMeta,
        key_id_map: KeyIdMap = None,
        skip_corrupted: Optional[bool] = False,
    ) -> VideoAnnotation:
        """
        Convert a json dict to VideoAnnotation. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Dict in json format.
        :type data: dict
        :param project_meta: Input ProjectMeta object.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: Union[:class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, None]
        :param skip_corrupted: Skip corrupted items (currently only frames) during conversion.
        :type skip_corrupted: bool, optional
        :returns: VideoAnnotation object
        :rtype: :class:`~supervisely.video_annotation.video_annotation.VideoAnnotation`

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                video_ann_json = {
                    "size": {"height": 500, "width": 700},
                    "tags": [],
                    "objects": [],
                    "frames": [],
                    "framesCount": 1
                }
                key_id_map = sly.KeyIdMap()
                meta = sly.ProjectMeta()

                video_ann = sly.VideoAnnotation.from_json(video_ann_json, meta, key_id_map)
        """
        # video_name = data[VIDEO_NAME]
        video_key = uuid.UUID(data[KEY]) if KEY in data else uuid.uuid4()

        if key_id_map is not None:
            key_id_map.add_video(video_key, data.get(VIDEO_ID, None))

        img_size_dict = data[IMG_SIZE]
        img_height = img_size_dict[IMG_SIZE_HEIGHT]
        img_width = img_size_dict[IMG_SIZE_WIDTH]
        img_size = (img_height, img_width)

        description = data.get(DESCRIPTION, "")
        frames_count = data[FRAMES_COUNT]

        tags = VideoTagCollection.from_json(data[TAGS], project_meta.tag_metas, key_id_map)
        objects = VideoObjectCollection.from_json(data[OBJECTS], project_meta, key_id_map)
        frames = FrameCollection.from_json(
            data[FRAMES], objects, frames_count, key_id_map, skip_corrupted
        )

        return cls(
            img_size=img_size,
            frames_count=frames_count,
            objects=objects,
            frames=frames,
            tags=tags,
            description=description,
            key=video_key,
        )

    @classmethod
    def load_json_file(
        cls, path: str, project_meta: ProjectMeta, key_id_map: Optional[KeyIdMap] = None
    ) -> VideoAnnotation:
        """
        Loads json file and converts it to VideoAnnotation.

        :param path: Path to the json file.
        :type path: str
        :param project_meta: Input ProjectMeta object.
        :type project_meta: :class:`~supervisely.project.project_meta.ProjectMeta`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: VideoAnnotation object
        :rtype: :class:`~supervisely.video_annotation.video_annotation.VideoAnnotation`

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                team_name = 'Vehicle Detection'
                workspace_name = 'Cities'
                project_name =  'London'

                team = api.team.get_info_by_name(team_name)
                workspace = api.workspace.get_info_by_name(team.id, workspace_name)
                project = api.project.get_info_by_name(workspace.id, project_name)

                meta_json = api.project.get_meta(project.id)
                meta = sly.ProjectMeta.from_json(meta_json)

                # Load json file
                path = "/home/admin/work/docs/my_dataset/ann/annotation.json"
                ann = sly.VideoAnnotation.load_json_file(path, meta)
        """
        with open(path) as fin:
            data = json.load(fin)
        return cls.from_json(data, project_meta, key_id_map)

    def clone(
        self,
        img_size: Optional[Tuple[int, int]] = None,
        frames_count: Optional[int] = None,
        objects: Optional[VideoObjectCollection] = None,
        frames: Optional[FrameCollection] = None,
        tags: Optional[VideoTagCollection] = None,
        description: Optional[str] = None,
    ) -> VideoAnnotation:
        """
        Makes a copy of VideoAnnotation with new fields, if fields are given, otherwise it will use fields of the original VideoAnnotation.

        :param img_size: Size of the image (height, width).
        :type img_size: Tuple[int, int], optional
        :param frames_count: Number of frames in VideoAnnotation.
        :type frames_count: int, optional
        :param objects: VideoObjectCollection object.
        :type objects: :class:`~supervisely.video_annotation.video_object_collection.VideoObjectCollection`, optional
        :param frames: FrameCollection object.
        :type frames: :class:`~supervisely.video_annotation.frame_collection.FrameCollection`, optional
        :param tags: VideoTagCollection object.
        :type tags: :class:`~supervisely.video_annotation.video_tag_collection.VideoTagCollection`, optional
        :param description: Video description.
        :type description: str, optional
        :raises TypeError: if img_size is not tuple or list

        :Usage Example:

            .. code-block:: python

                import supervisely as sly

                height, width = 500, 700
                frames_count = 1
                video_ann = sly.VideoAnnotation((height, width), frames_count)

                obj_class_car = sly.ObjClass('car', sly.Rectangle)
                video_obj_car = sly.VideoObject(obj_class_car)
                new_objects = sly.VideoObjectCollection([video_obj_car])
                new_video_ann = video_ann.clone(objects=new_objects)
                print(new_video_ann.to_json())
                # Output: {
                #     "size": {
                #         "height": 500,
                #         "width": 700
                #     },
                #     "description": "",
                #     "key": "37f7d267864c4fd8b1a1a32f67e37f7d",
                #     "tags": [],
                #     "objects": [
                #         {
                #             "key": "27d4ba1aaee64930b2d0bfb7e8b53493",
                #             "classTitle": "car",
                #             "tags": []
                #         }
                #     ],
                #     "frames": [],
                #     "framesCount": 1
                # }
        """
        return VideoAnnotation(
            img_size=take_with_default(img_size, self.img_size),
            frames_count=take_with_default(frames_count, self.frames_count),
            objects=take_with_default(objects, self.objects),
            frames=take_with_default(frames, self.frames),
            tags=take_with_default(tags, self.tags),
            description=take_with_default(description, self.description),
        )

    def is_empty(self) -> bool:
        """
        Check whether video annotation contains objects or tags, or not.

        :returns: True if video annotation is empty, False otherwise.
        :rtype: bool

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.video_annotation.key_id_map import KeyIdMap

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 17208
                video_id = 19371139
                key_id_map = KeyIdMap()
                meta_json = api.project.get_meta(project_id)
                meta = sly.ProjectMeta.from_json(meta_json)

                ann_json = api.video.annotation.download(video_id)
                ann = sly.VideoAnnotation.from_json(ann_json, meta, key_id_map)

                print(ann.is_empty()) # False
        """

        if len(self.objects) == 0 and len(self.tags) == 0:
            return True
        else:
            return False



# coding: utf-8
from collections.abc import Mapping

from supervisely.io.json import load_json_file

from jsonschema import Draft4Validator, validators


# From https://python-jsonschema.readthedocs.io/en/v2.6.0/faq/
def _extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for prop, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(prop, subschema["default"])

        for error in validate_properties(validator, properties, instance, schema):
            yield error

    return validators.extend(validator_class, {"properties": set_defaults})


class MultiTypeValidator(object):
    """JSON schema validator that holds multiple named validators loaded from a single schema file."""

    DEFINITIONS = 'definitions'

    def __init__(self, schema_fpath):
        """
        :param schema_fpath: Path to JSON schema file.
        :type schema_fpath: str
        """
        vtor_class = _extend_with_default(Draft4Validator)
        schemas = load_json_file(schema_fpath)
        # Detach common definitions from the named schemas and inline them in into every schema.
        definitions = schemas.pop(self.DEFINITIONS, {})
        for name, schema in schemas.items():
            schema.setdefault(self.DEFINITIONS, {}).update(definitions)

        self._validators = {name: vtor_class(schema) for name, schema in schemas.items()}

    def val(self, type_name, obj):
        validator = self._validators.get(type_name)
        if validator is None:
            raise RuntimeError('JSON validator is not defined. Type: {}'.format(type_name))
        try:
            validator.validate(obj)
        except Exception as e:
            raise RuntimeError('Error occurred during JSON validation. Type: {}. Exc: {}. See documentation'.format(
                type_name, str(e)
            )) from None  # suppress previous stacktrace, save all required info


class JsonConfigValidator(MultiTypeValidator):
    """Validator for training and inference configs based on bundled JSON schemas."""

    def __init__(self, schema_fpath=None):
        """
        :param schema_fpath: Path to schemas.json or None for default.
        :type schema_fpath: str
        """
        super().__init__(schema_fpath or '/workdir/src/schemas.json')

    def validate_train_cfg(self, config):
        # store all possible requirements in schema, including size % 16 etc
        self.val('training_config', config)

    def validate_inference_cfg(self, config):
        # store all possible requirements in schema
        self.val('inference_config', config)


class AlwaysPassingConfigValidator:
    """Validator stub that accepts any config (used when schema validation is unavailable/disabled)."""

    @classmethod
    def validate_train_cfg(cls, config):
        pass

    @classmethod
    def validate_inference_cfg(cls, config):
        pass


def update_recursively(lhs, rhs):
    """Performs updating like lhs.update(rhs), but operates recursively on nested dictionaries."""
    for k, v in rhs.items():
        if isinstance(v, Mapping):
            lhs[k] = update_recursively(lhs.get(k, {}), v)
        else:
            lhs[k] = v
    return lhs


def update_strict(lhs, rhs):
    for k, v in rhs.items():
        if k in lhs:
            raise ValueError('Key collision while strict updating. Key:"{}"'.format(str(k)))
        lhs[k] = v
    return lhs



import functools
import inspect
import json
import traceback
from typing import Any, BinaryIO, Dict, List, Optional, Tuple, Union

from fastapi import Form, Request, Response, UploadFile, status
from pydantic import ValidationError

from supervisely._utils import find_value_by_keys
from supervisely.api.api import Api
from supervisely.api.module_api import ApiField
from supervisely.io import env
from supervisely.nn.inference.inference import (
    Inference,
    _get_log_extra_for_inference_request,
)
from supervisely.sly_logger import logger


def validate_key(data: Dict, key: str, type_: type):
    if key not in data:
        raise ValidationError(f"Key {key} not found in inference request.")
    if not isinstance(data[key], type_):
        raise ValidationError(f"Key {key} is not of type {type_}.")


def handle_validation(func):
    def _find_response(args, kwargs):
        for arg in args:
            if isinstance(arg, Response):
                return arg
        for value in kwargs.values():
            if isinstance(value, Response):
                return value
        return None

    def _handle_exception(e, response):
        if response is not None:
            logger.error(f"ValidationError: {e}", exc_info=True)
            response.status_code = status.HTTP_400_BAD_REQUEST
            return {"error": str(e), "success": False}
        raise e

    if inspect.iscoroutinefunction(func):

        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            response = _find_response(args, kwargs)
            try:
                return await func(*args, **kwargs)
            except ValidationError as e:
                return _handle_exception(e, response)

        return async_wrapper

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        response = _find_response(args, kwargs)
        try:
            return func(*args, **kwargs)
        except ValidationError as e:
            return _handle_exception(e, response)

    return wrapper


class BaseTracking(Inference):
    """Base for video object tracking: propagates annotations frame-to-frame via model or heuristic."""

    def __init__(
        self,
        model_dir: Optional[str] = None,
        custom_inference_settings: Optional[Union[Dict[str, Any], str]] = None,
    ):
        """See :class:`~supervisely.nn.inference.inference.Inference` for params."""
        Inference.__init__(
            self,
            model_dir,
            custom_inference_settings,
            sliding_window_mode=None,
            use_gui=False,
        )

        try:
            self.load_on_device(model_dir, "cuda")
        except RuntimeError:
            self.load_on_device(model_dir, "cpu")
            logger.warning("Failed to load model on CUDA device.")

        logger.debug(
            "Smart cache params",
            extra={"ttl": env.smart_cache_ttl(), "maxsize": env.smart_cache_size()},
        )

    def get_info(self):
        info = super().get_info()
        info["task type"] = "tracking"
        return info

    @staticmethod
    def _notify_error_default(
        api: Api, track_id: str, exception: Exception, with_traceback: bool = False
    ):
        error_name = type(exception).__name__
        message = str(exception)
        if with_traceback:
            message = f"{message}\n{traceback.format_exc()}"
        api.video.notify_tracking_error(track_id, error_name, message)

    @staticmethod
    def _notify_error_direct(
        api: Api,
        session_id: str,
        video_id,
        track_id: str,
        exception: Exception,
        with_traceback: bool = False,
    ):
        error_name = type(exception).__name__
        message = str(exception)
        if with_traceback:
            message = f"{message}\n{traceback.format_exc()}"
        api.vid_ann_tool.set_direct_tracking_error(
            session_id=session_id,
            video_id=video_id,
            track_id=track_id,
            message=f"{error_name}: {message}",
        )

    @staticmethod
    def send_error_data(api, context):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                try:
                    return func(*args, **kwargs)
                except Exception as exc:
                    try:
                        track_id = context["trackId"]
                        if ApiField.USE_DIRECT_PROGRESS_MESSAGES in context:
                            session_id = find_value_by_keys(context, ["sessionId", "session_id"])
                            video_id = find_value_by_keys(context, ["videoId", "video_id"])
                            BaseTracking._notify_error_direct(
                                api=api,
                                session_id=session_id,
                                video_id=video_id,
                                track_id=track_id,
                                exception=exc,
                                with_traceback=False,
                            )
                        else:
                            BaseTracking._notify_error_default(
                                api=api, track_id=track_id, exception=exc, with_traceback=False
                            )
                    except Exception:
                        logger.error("An error occurred while sending error data", exc_info=True)
                    raise exc

            return wrapper

        return decorator

    def _pop_tracking_results(self, inference_request_uuid: str, frame_range: Tuple = None):
        inference_request = self.inference_requests_manager.get(inference_request_uuid)
        logger.debug(
            "Pop tracking results",
            extra={
                "inference_request_uuid": inference_request_uuid,
                "pending_results_len": inference_request.pending_num(),
                "frame_range": frame_range,
            },
        )

        data = {}
        if frame_range is not None:

            def _in_range(figure):
                return figure.frame_index >= frame_range[0] and figure.frame_index <= frame_range[1]

            with inference_request._lock:
                data["pending_results"] = [
                    x for x in inference_request._pending_results if _in_range(x)
                ]
                inference_request._pending_results = [
                    x for x in inference_request._pending_results if not _in_range(x)
                ]
        else:
            data["pending_results"] = inference_request.pop_pending_results()

        data = {
            **inference_request.to_json(),
            **_get_log_extra_for_inference_request(inference_request.uuid, inference_request),
            "pending_results": data["pending_results"],
        }

        return data

    def _clear_tracking_results(self, inference_request_uuid):
        if inference_request_uuid is None:
            raise ValueError("'inference_request_uuid' is required.")
        self.inference_requests_manager.remove_after(inference_request_uuid, 60)
        logger.debug("Removed an inference request:", extra={"uuid": inference_request_uuid})
        return {"success": True}

    def _stop_tracking(self, inference_request_uuid: str):
        inference_request = self.inference_requests_manager.get(inference_request_uuid)
        inference_request.stop()
        logger.debug(
            "Stopped tracking:",
            extra={
                "uuid": inference_request_uuid,
                "inference_request_uuid": inference_request_uuid,
            },
        )

    # Implement the following methods in the derived class
    def track(self, api: Api, state: Dict, context: Dict):
        raise NotImplementedError("Method `track` must be implemented.")

    def track_api(self, api: Api, state: Dict, context: Dict):
        raise NotImplementedError("Method `_track_api` must be implemented.")

    def track_api_files(
        self,
        files: List[BinaryIO],
        settings: Dict,
    ):
        raise NotImplementedError("Method `track_api_files` must be implemented.")

    def track_async(self, api: Api, state: Dict, context: Dict):
        raise NotImplementedError("Method `track_async` must be implemented.")

    def stop_tracking(self, state: Dict, context: Dict):
        validate_key(context, "inference_request_uuid", str)
        inference_request_uuid = context["inference_request_uuid"]
        self._stop_tracking(inference_request_uuid)
        return {"message": "Inference will be stopped.", "success": True}

    def pop_tracking_results(self, state: Dict, context: Dict):
        validate_key(context, "inference_request_uuid", str)
        inference_request_uuid = context["inference_request_uuid"]

        inference_request = self.inference_requests_manager.get(inference_request_uuid)
        log_extra = _get_log_extra_for_inference_request(inference_request.uuid, inference_request)
        frame_range = find_value_by_keys(context, ["frameRange", "frame_range", "frames"])
        tracking_results = self._pop_tracking_results(inference_request_uuid, frame_range)
        logger.debug(f"Sending inference delta results with uuid:", extra=log_extra)
        return tracking_results

    def clear_tracking_results(self, state: Dict, context: Dict):
        inference_request_uuid = context.get("inference_request_uuid", None)
        self._clear_tracking_results(inference_request_uuid)
        return {"message": "Inference results cleared.", "success": True}

    def _register_endpoints(self):
        server = self._app.get_server()

        @server.post("/track")
        @handle_validation
        def track_handler(request: Request):
            api = self.api_from_request(request)
            state = request.state.state
            context = request.state.context
            logger.info("Received track request.", extra={"context": context, "state": state})
            return self.track(api, state, context)

        @server.post("/track-api")
        @handle_validation
        async def track_api_handler(request: Request):
            api = self.api_from_request(request)
            state = request.state.state
            context = request.state.context
            logger.info("Received track-api request.", extra={"context": context, "state": state})
            result = self.track_api(api, state, context)
            logger.info("Track-api request processed.")
            return result

        @server.post("/track-api-files")
        @handle_validation
        def track_api_files(
            files: List[UploadFile],
            settings: str = Form("{}"),
        ):
            files = [file.file for file in files]
            settings = json.loads(settings)
            return self.track_api_files(files, settings)

        @server.post("/track_async")
        @handle_validation
        def track_async_handler(request: Request):
            api = self.api_from_request(request)
            state = request.state.state
            context = request.state.context
            logger.info("Received track_async request.", extra={"context": context, "state": state})
            return self.track_async(api, state, context)

        @server.post("/stop_tracking")
        @handle_validation
        def stop_tracking_handler(response: Response, request: Request):
            state = request.state.state
            context = request.state.context
            logger.info(
                "Received stop_tracking request.", extra={"context": context, "state": state}
            )
            return self.stop_tracking(state, context)

        @server.post("/pop_tracking_results")
        @handle_validation
        def pop_tracking_results_handler(request: Request, response: Response):
            state = request.state.state
            context = request.state.context
            logger.info(
                "Received pop_tracking_results request.", extra={"context": context, "state": state}
            )
            return self.pop_tracking_results(state, context)

        @server.post("/clear_tracking_results")
        @handle_validation
        def clear_tracking_results_handler(request: Request, response: Response):
            context = request.state.context
            state = request.state.state
            logger.info(
                "Received clear_tracking_results request.",
                extra={"context": context, "state": state},
            )
            return self.clear_tracking_results(state, context)

    def serve(self):
        super().serve()
        self._register_endpoints()



# coding: utf-8
"""Work with video objects via the Supervisely API."""

# docs
from typing import List, Optional

from supervisely.api.entity_annotation.object_api import ObjectApi
from supervisely.api.video.video_tag_api import VideoObjectTagApi
from supervisely.video_annotation.key_id_map import KeyIdMap
from supervisely.video_annotation.video_object_collection import VideoObjectCollection


class VideoObjectApi(ObjectApi):
    """
    API for working with video objects.
    """

    def __init__(self, api):
        """
        :param api: :class:`~supervisely.api.api.Api` object to use for API connection.
        :type api: :class:`~supervisely.api.api.Api`
        """
        super().__init__(api)
        self.tag = VideoObjectTagApi(api)

    def append_bulk(
        self,
        video_id: int,
        objects: VideoObjectCollection,
        key_id_map: Optional[KeyIdMap] = None,
    ) -> List[int]:
        """
        Add objects to annotation objects.

        :param video_id: Video ID in Supervisely.
        :type video_id: int
        :param objects: VideoObjectCollection objects collection.
        :type objects: :class:`~supervisely.video_annotation.video_object_collection.VideoObjectCollection`
        :param key_id_map: KeyIdMap object.
        :type key_id_map: :class:`~supervisely.video_annotation.key_id_map.KeyIdMap`, optional
        :returns: List of object IDs
        :rtype: List[int]

        :Usage Example:

            .. code-block:: python

                import os
                from dotenv import load_dotenv

                import supervisely as sly
                from supervisely.video_annotation.key_id_map import KeyIdMap

                # Load secrets and create API object from .env file (recommended)
                # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
                if sly.is_development():
                    load_dotenv(os.path.expanduser("~/supervisely.env"))

                api = sly.Api.from_env()

                project_id = 17209
                video_id = 19402023

                meta_json = api.project.get_meta(project_id)
                project_meta = sly.ProjectMeta.from_json(meta_json)

                key_id_map = KeyIdMap()
                ann_info = api.video.annotation.download(video_id)
                ann = sly.VideoAnnotation.from_json(ann_info, project_meta, key_id_map)

                api.video.object.append_bulk(video_id, ann.objects, key_id_map)
        """

        info = self._api.video.get_info_by_id(video_id)
        return self._append_bulk(
            self._api.video.tag, video_id, info.project_id, info.dataset_id, objects, key_id_map
        )



# coding: utf-8
from __future__ import annotations

import warnings
from copy import deepcopy
from math import ceil, floor
from typing import TYPE_CHECKING, Dict, List, Tuple

import numpy as np

from supervisely.sly_logger import logger
from supervisely.geometry.constants import (
    ANY_SHAPE,
    BITMAP,
    CLASS_ID,
    CREATED_AT,
    EXTERIOR,
    ID,
    INTERIOR,
    LABELER_LOGIN,
    LOC,
    NODES,
    ORIGIN,
    POINTS,
    UPDATED_AT,
)
from supervisely.io.json import JsonSerializable

warnings.simplefilter(action="ignore", category=FutureWarning)

if not hasattr(np, "bool"):
    np.bool = np.bool_

if TYPE_CHECKING:
    from supervisely.geometry.rectangle import Rectangle
    from supervisely.geometry.image_rotator import ImageRotator


# @TODO: use properties instead of field if it makes sense
class Geometry(JsonSerializable):
    """Base class for all geometry classes."""

    def __init__(
        self,
        sly_id=None,
        class_id=None,
        labeler_login=None,
        updated_at=None,
        created_at=None,
    ):
        """
        :param sly_id: Geometry ID on Supervisely server.
        :type sly_id: int, optional
        :param class_id: ObjClass ID.
        :type class_id: int, optional
        :param labeler_login: Labeler user login.
        :type labeler_login: str, optional
        :param updated_at: Last update timestamp.
        :type updated_at: int, optional
        :param created_at: Creation timestamp.
        :type created_at: int, optional
        """
        self.sly_id = sly_id
        self.labeler_login = labeler_login
        self.updated_at = updated_at
        self.created_at = created_at
        self.class_id = class_id

    def _add_creation_info(self, d):
        """Add creation information to the geometry."""
        if self.labeler_login is not None:
            d[LABELER_LOGIN] = self.labeler_login
        if self.updated_at is not None:
            d[UPDATED_AT] = self.updated_at
        if self.created_at is not None:
            d[CREATED_AT] = self.created_at
        if self.sly_id is not None:
            d[ID] = self.sly_id
        if self.class_id is not None:
            d[CLASS_ID] = self.class_id

    def _copy_creation_info_inplace(self, g):
        """Copy creation information to the geometry."""
        self.labeler_login = g.labeler_login
        self.updated_at = g.updated_at
        self.created_at = g.created_at
        self.sly_id = g.sly_id

    @staticmethod
    def geometry_name():
        """
        Get the name of the geometry.

        :returns: name of the geometry
        :rtype: str
        :returns: string with name of geometry
        """
        raise NotImplementedError()

    @classmethod
    def name(cls):
        """
        Get the name of the geometry.

        Same as :meth:`~supervisely.geometry.geometry.Geometry.geometry_name`, but shorter. In order to make the code more concise.

        :returns: string with name of geometry
        """
        return cls.geometry_name()

    def crop(self, rect: Rectangle) -> List[Geometry]:
        """
        Crop the geometry with given rectangle.

        :param rect: Rectangle for crop.
        :type rect: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: List of Geometry after crop.
        :rtype: List[:class:`~supervisely.geometry.geometry.Geometry`]`
        """
        raise NotImplementedError()

    def relative_crop(self, rect):
        """Crops object like "crop" method, but return results with coordinates relative to rect
        :param rect: Rectangle for crop.
        :type rect: :class:`~supervisely.geometry.rectangle.Rectangle`
        :returns: List of Geometry after relative crop.
        :rtype: List[:class:`~supervisely.geometry.geometry.Geometry`]`
        :raises NotImplementedError: if method is not implemented in subclass
        """
        return [geom.translate(drow=-rect.top, dcol=-rect.left) for geom in self.crop(rect)]

    def rotate(self, rotator: ImageRotator) -> Geometry:
        """Rotates around image center -> New Geometry
        :param rotator: Class for image rotation.
        :type rotator: :class:`~supervisely.geometry.image_rotator.ImageRotator`
        :returns: Geometry after rotation.
        :rtype: :class:`~supervisely.geometry.geometry.Geometry`
        """
        raise NotImplementedError()

    def resize(self, in_size: Tuple[int, int], out_size: Tuple[int, int]) -> Geometry:
        """
        :param in_size: Input image size (height, width).
        :type in_size: Tuple[int, int]
        :param out_size: Desired output image size (height, width) of the Annotation to which Label belongs.
        :type out_size: Tuple[int, int]
        :returns: Geometry after resize.
        :rtype: :class:`~supervisely.geometry.geometry.Geometry`
        :raises NotImplementedError: if method is not implemented in subclass
        """
        raise NotImplementedError()

    def scale(self, factor):
        """Scales around origin with a given factor.
        :param: factor: Scale factor.
        :type factor: float
        :returns: :class:`~supervisely.geometry.geometry.Geometry`
        :rtype: :class:`~supervisely.geometry.geometry.Geometry`
        :raises NotImplementedError: if method is not implemented in subclass
        """
        raise NotImplementedError()

    def translate(self, drow, dcol):
        """
        :param drow: Vertical shift.
        :type drow: int
        :param dcol: Horizontal shift.
        :type dcol: int
        :returns: Geometry after translate.
        :rtype: :class:`~supervisely.geometry.geometry.Geometry`
        :raises NotImplementedError: if method is not implemented in subclass
        """
        raise NotImplementedError()

    def fliplr(self, img_size):
        """
        :param img_size: Image size (height, width) to which belongs Geometry.
        :type img_size: Tuple[int, int]
        :returns: Geometry after flip in horizontal.
        :rtype: :class:`~supervisely.geometry.geometry.Geometry`
        :raises NotImplementedError: if method is not implemented in subclass
        """
        raise NotImplementedError()

    def flipud(self, img_size):
        """
        :param img_size: Image size (height, width) to which belongs Geometry.
        :type img_size: Tuple[int, int]
        :returns: Geometry after flip in vertical.
        :rtype: :class:`~supervisely.geometry.geometry.Geometry`
        :raises NotImplementedError: if method is not implemented in subclass
        """
        raise NotImplementedError()

    def _draw_bool_compatible(self, draw_fn, bitmap, color, thickness, config=None):
        """ """
        if bitmap.dtype == np.bool:
            # Cannot draw on the canvas directly, create a temporary with different type.
            temp_bitmap = np.zeros(bitmap.shape[:2], dtype=np.uint8)
            draw_fn(temp_bitmap, 1, thickness=thickness, config=config)
            bitmap[temp_bitmap == 1] = color
        else:
            # Pass through the canvas without temp bitmap for efficiency.
            draw_fn(bitmap, color, thickness=thickness, config=config)

    def draw(self, bitmap, color, thickness=1, config=None):
        """
        :param bitmap: np.ndarray
        :param color: Color [R, G, B]
        :param thickness: used only in Polyline and :class:`~supervisely.geometry.point.Point`
        :param config: Drawing config specific to a concrete subclass, e.g. per edge colors
        :type config: dict
        :raises NotImplementedError: if method is not implemented in subclass
        """
        self._draw_bool_compatible(self._draw_impl, bitmap, color, thickness, config)

    def get_mask(self, img_size: Tuple[int, int]):
        """Returns 2D boolean mask of the geometry.
        With shape as img_size (height, width) and filled
        with True values inside the geometry and False values outside.
        dtype = np.bool
        shape = img_size

        :param img_size: size of the image (height, width)
        :type img_size: Tuple[int, int]
        :returns: 2D boolean mask of the geometry
        :rtype: np.ndarray
        """
        bitmap = np.zeros(img_size + (3,), dtype=np.uint8)
        self.draw(bitmap, color=[255, 255, 255], thickness=-1)
        return np.any(bitmap != 0, axis=-1)

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        """
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: used only in Polyline and :class:`~supervisely.geometry.point.Point`
        :type thickness: int
        :param config: Drawing config specific to a concrete subclass, e.g. per edge colors
        :type config: dict
        :raises NotImplementedError: if method is not implemented in subclass
        """
        raise NotImplementedError()

    def draw_contour(self, bitmap, color, thickness=1, config=None):
        """Draws the figure contour on a given bitmap canvas
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: (int)
        :param config: drawing config specific to a concrete subclass, e.g. per edge colors
        """
        self._draw_bool_compatible(self._draw_contour_impl, bitmap, color, thickness, config)

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """Draws the figure contour on a given bitmap canvas
        :param bitmap: np.ndarray
        :param color: [R, G, B]
        :param thickness: (int)
        """
        raise NotImplementedError()

    @property
    def area(self):
        """
        :returns: float
        """
        raise NotImplementedError()

    def to_bbox(self) -> Rectangle:
        """
        :returns: Rectangle from geometry.
        :rtype: :class:`~supervisely.geometry.rectangle.Rectangle`
        :raises NotImplementedError: if method is not implemented in subclass
        """
        raise NotImplementedError()

    def clone(self):
        """Clone from GEOMETRYYY"""
        return deepcopy(self)

    def validate(self, obj_class_shape, settings):
        """
        Validate geometry.

        :param obj_class_shape: Object class shape.
        :type obj_class_shape: str
        :param settings: Settings.
        :type settings: dict
        :raises ValueError: if geometry validation error
        """
        if obj_class_shape != ANY_SHAPE:
            if self.geometry_name() != obj_class_shape:
                raise ValueError("Geometry validation error: shape names are mismatched!")

    @staticmethod
    def config_from_json(config):
        """
        Convert geometry config from json format

        :param config: dictionary(geometry config) in json format
        :type config: dict
        :returns: dictionary(geometry config) in json format
        :rtype: dict
        """
        return config

    @staticmethod
    def config_to_json(config):
        """
        Convert geometry config to json format

        :param config: dictionary(geometry config)
        :type config: dict
        :returns: dictionary(geometry config) in json format
        :rtype: dict
        """
        return config

    @classmethod
    def allowed_transforms(cls):
        """
        Returns the allowed transforms for the Geometry.
        """
        # raise NotImplementedError("{!r}".format(cls.geometry_name()))
        return []

    def convert(self, new_geometry, contour_radius=0, approx_epsilon=None):
        """
        Convert geometry to another geometry shape.

        :param new_geometry: New geometry shape.
        :type new_geometry: :class:`~supervisely.geometry.geometry.Geometry`
        :param contour_radius: Radius of the contour.
        :type contour_radius: int
        :param approx_epsilon: Approximation epsilon.
        :type approx_epsilon: float
        :returns: List of geometries.
        :rtype: List[:class:`~supervisely.geometry.geometry.Geometry`]
        """
        from supervisely.geometry.any_geometry import AnyGeometry

        if type(self) == new_geometry or new_geometry == AnyGeometry:
            return [self]

        allowed_transforms = self.allowed_transforms()
        if new_geometry not in allowed_transforms:
            raise NotImplementedError(
                "from {!r} to {!r}".format(self.geometry_name(), new_geometry.geometry_name())
            )

        from supervisely.geometry.alpha_mask import AlphaMask
        from supervisely.geometry.bitmap import Bitmap
        from supervisely.geometry.helpers import (
            geometry_to_alpha_mask,
            geometry_to_bitmap,
            geometry_to_polygon,
        )
        from supervisely.geometry.polygon import Polygon
        from supervisely.geometry.rectangle import Rectangle
        from supervisely.geometry.oriented_bbox import OrientedBBox

        res = []
        if new_geometry == Bitmap:
            res = geometry_to_bitmap(self, radius=contour_radius)
        elif new_geometry == AlphaMask:
            res = geometry_to_alpha_mask(self, radius=contour_radius)
        elif new_geometry == Rectangle:
            res = [self.to_bbox()]
        elif new_geometry == Polygon:
            res = geometry_to_polygon(self, approx_epsilon=approx_epsilon)
        elif new_geometry == OrientedBBox:
            bbox = self.to_bbox()
            res = [OrientedBBox.from_bbox(bbox)]

        if len(res) == 0:
            logger.warn(
                "Can not convert geometry {} to {} because geometry to convert is very small".format(
                    self.geometry_name(), new_geometry.geometry_name()
                )
            )
        return res

    @classmethod
    def _to_pixel_coordinate_system_json(cls, data: Dict, image_size: List[int]) -> Dict:
        """
        Convert geometry from subpixel precision to pixel precision by subtracting a subpixel offset from the coordinates.

        This method should be reimplemented in subclasses if needed.

        Point order: [x, y]

        In the labeling tool, labels are created with subpixel precision,
        which means that the coordinates of the geometry can have decimal values representing fractions of a pixel.
        However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

        :param data: Json data with geometry config.
        :type data: dict
        :param image_size: Image size in pixels (height, width).
        :type image_size: List[int]
        :returns: Json data with coordinates converted to pixel coordinate system.
        :rtype: dict
        """
        data = deepcopy(data)  # Avoid modifying the original data
        height, width = image_size[:2]

        # Point, Polygon, Polyline. Rectangle have its own implementation
        if data.get(POINTS) is not None:
            exterior = data[POINTS][EXTERIOR]
            interior = data[POINTS][INTERIOR]
            for point in exterior:
                point[0] = floor(point[0]) - 1 if point[0] == width else floor(point[0])
                point[1] = floor(point[1]) - 1 if point[1] == height else floor(point[1])
            for coords in interior:
                for point in coords:
                    point[0] = floor(point[0]) - 1 if point[0] == width else floor(point[0])
                    point[1] = floor(point[1]) - 1 if point[1] == height else floor(point[1])
            data[POINTS][EXTERIOR] = exterior
            data[POINTS][INTERIOR] = interior

        # Bitmap and AlphaMask
        if data.get(BITMAP) is not None:
            origin = data[BITMAP][ORIGIN]
            data[BITMAP][ORIGIN] = [floor(origin[0]), floor(origin[1])]

        # GraphNodes and Cuboid
        if data.get(NODES) is not None:
            nodes = data[NODES]
            for node_key in nodes:
                nodes[node_key][LOC] = [
                    (
                        floor(nodes[node_key][LOC][0]) - 1
                        if nodes[node_key][LOC][0] == width
                        else floor(nodes[node_key][LOC][0])
                    ),
                    (
                        floor(nodes[node_key][LOC][1]) - 1
                        if nodes[node_key][LOC][1] == height
                        else floor(nodes[node_key][LOC][1])
                    ),
                ]
            data[NODES] = nodes
        return data

    @classmethod
    def _to_subpixel_coordinate_system_json(cls, data: Dict) -> Dict:
        """
        Convert geometry from pixel precision to subpixel precision by adding a subpixel offset to the coordinates.

        This method should be reimplemented in subclasses if needed.

        Point order: [x, y]

        In the labeling tool, labels are created with subpixel precision,
        which means that the coordinates of the geometry can have decimal values representing fractions of a pixel.
        However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

        :param data: Json data with geometry config.
        :type data: dict
        :returns: Json data with coordinates converted to subpixel coordinate system.
        :rtype: dict
        """
        data = deepcopy(data)  # Avoid modifying the original data

        # Point, Polygon, Polyline. Rectangle have its own implementation
        if data.get(POINTS) is not None:
            exterior = data[POINTS][EXTERIOR]
            interior = data[POINTS][INTERIOR]
            for point in exterior:
                point[0] = point[0] + 0.5
                point[1] = point[1] + 0.5
            for coords in interior:
                for point in coords:
                    point[0] = point[0] + 0.5
                    point[1] = point[1] + 0.5
            data[POINTS][EXTERIOR] = exterior
            data[POINTS][INTERIOR] = interior

        # Bitmap and AlphaMask
        if data.get(BITMAP) is not None:
            origin = data[BITMAP][ORIGIN]
            data[BITMAP][ORIGIN] = [floor(origin[0]), floor(origin[1])]

        # GraphNodes and Cuboid
        if data.get(NODES) is not None:
            nodes = data[NODES]
            for node_key in nodes:
                nodes[node_key][LOC] = [
                    floor(nodes[node_key][LOC][0]) + 0.5,
                    floor(nodes[node_key][LOC][1]) + 0.5,
                ]
            data[NODES] = nodes
        return data

    # def _to_pixel_coordinate_system(self):
    #     """
    #     This method should be implemented in subclasses.

    #     Convert geometry from subpixel precision to pixel precision by subtracting a subpixel offset from the coordinates.

    #     In the labeling tool, labels are created with subpixel precision,
    #     which means that the coordinates of the geometry can have decimal values representing fractions of a pixel.
    #     However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

    #     :returns: New instance of Geometry object in pixel coordinates system
    #     :rtype: :class:`~supervisely.geometry.geometry.Geometry`
    #     """
    #     return self

    # def _to_subpixel_coordinate_system(self):
    #     """
    #     This method should be implemented in subclasses.

    #     Convert geometry from pixel precision to subpixel precision by adding a subpixel offset to the coordinates.

    #     In the labeling tool, labels are created with subpixel precision,
    #     which means that the coordinates of the geometry can have decimal values representing fractions of a pixel.
    #     However, in Supervisely SDK, geometry coordinates are represented using pixel precision, where the coordinates are integers representing whole pixels.

    #     :returns: New instance of Geometry object in subpixel coordinates system
    #     :rtype: :class:`~supervisely.geometry.geometry.Geometry`
    #     """
    #     return self



# coding: utf-8

from supervisely.geometry.constants import LABELER_LOGIN, UPDATED_AT, CREATED_AT, ID, CLASS_ID, INDICES
from supervisely.geometry.geometry import Geometry

# pointcloud mask (segmentation)
class Pointcloud(Geometry):
    """3D point cloud geometry: reference to external point cloud file. Immutable."""

    @staticmethod
    def geometry_name():
        """
        Returns the name of the geometry.

        :returns: name of the geometry
        :rtype: str
        """
        return 'point_cloud'

    def __init__(self, indices, sly_id=None, class_id=None, labeler_login=None, updated_at=None, created_at=None):
        """Pointcloud initialization.

        :param indices: Indices of the pointcloud.
        :type indices: list
        :param sly_id: Pointcloud ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of ObjClass to which Pointcloud belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created Pointcloud.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Pointcloud was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Pointcloud was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional
        """
        super().__init__(sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)

        if type(indices) is not list:
            raise TypeError("\"indices\" param has to be of type {!r}".format(type(list)))

        self._indices = indices

    @property
    def indices(self):
        """
        Copy of the indices of the Pointcloud.

        :returns: indices of the :class:`~supervisely.geometry.pointcloud.Pointcloud`
        :rtype: list
        """
        return self._indices.copy()

    def to_json(self):
        """
        Converts the Pointcloud to a JSON object.

        :returns: JSON object
        :rtype: dict
        :returns: Pointcloud in json format
        """
        res = {INDICES: self.indices}
        self._add_creation_info(res)
        return res

    @classmethod
    def from_json(cls, data):
        """
        Converts a JSON object to a Pointcloud.

        :param data: JSON object
        :type data: dict
        :returns: Pointcloud
        :rtype: :class:`~supervisely.geometry.pointcloud.Pointcloud`
        :returns: Pointcloud from json.
        :rtype: :class:`~supervisely.geometry.pointcloud.Pointcloud`
        """
        indices = data[INDICES]

        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(indices, sly_id=sly_id, class_id=class_id, labeler_login=labeler_login, updated_at=updated_at, created_at=created_at)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc  # pylint: disable=import-error

import worker_api_pb2 as worker__api__pb2  # pylint: disable=import-error


class GeneralAPIStub(object):
    """package supervisely;"""

    def __init__(self, channel):
        """Initialize the GeneralAPIStub.

        :param channel: A grpc.Channel.
        :type channel: grpc.Channel
        """
        self.CreateProject = channel.unary_unary(
            "/GeneralAPI/CreateProject",
            request_serializer=worker__api__pb2.Project.SerializeToString,
            response_deserializer=worker__api__pb2.Id.FromString,
        )
        self.CreateDataset = channel.unary_unary(
            "/GeneralAPI/CreateDataset",
            request_serializer=worker__api__pb2.ProjectDataset.SerializeToString,
            response_deserializer=worker__api__pb2.Id.FromString,
        )
        self.GetProjectByName = channel.unary_unary(
            "/GeneralAPI/GetProjectByName",
            request_serializer=worker__api__pb2.Project.SerializeToString,
            response_deserializer=worker__api__pb2.Project.FromString,
        )
        self.GetDatasetByName = channel.unary_unary(
            "/GeneralAPI/GetDatasetByName",
            request_serializer=worker__api__pb2.ProjectDataset.SerializeToString,
            response_deserializer=worker__api__pb2.Dataset.FromString,
        )
        self.GetModelByName = channel.unary_unary(
            "/GeneralAPI/GetModelByName",
            request_serializer=worker__api__pb2.ModelName.SerializeToString,
            response_deserializer=worker__api__pb2.ModelInfo.FromString,
        )
        self.GetProjectMeta = channel.unary_unary(
            "/GeneralAPI/GetProjectMeta",
            request_serializer=worker__api__pb2.Id.SerializeToString,
            response_deserializer=worker__api__pb2.Project.FromString,
        )
        self.GetProjectDatasets = channel.unary_unary(
            "/GeneralAPI/GetProjectDatasets",
            request_serializer=worker__api__pb2.Id.SerializeToString,
            response_deserializer=worker__api__pb2.DatasetArray.FromString,
        )
        self.GetDatasetImages = channel.unary_unary(
            "/GeneralAPI/GetDatasetImages",
            request_serializer=worker__api__pb2.Id.SerializeToString,
            response_deserializer=worker__api__pb2.ImageArray.FromString,
        )
        self.GetImagesInfo = channel.unary_unary(
            "/GeneralAPI/GetImagesInfo",
            request_serializer=worker__api__pb2.ImageArray.SerializeToString,
            response_deserializer=worker__api__pb2.ImagesInfo.FromString,
        )
        self.DownloadImages = channel.unary_stream(
            "/GeneralAPI/DownloadImages",
            request_serializer=worker__api__pb2.ImagesHashes.SerializeToString,
            response_deserializer=worker__api__pb2.ChunkImage.FromString,
        )
        self.DownloadAnnotations = channel.unary_stream(
            "/GeneralAPI/DownloadAnnotations",
            request_serializer=worker__api__pb2.ImageArray.SerializeToString,
            response_deserializer=worker__api__pb2.ChunkImage.FromString,
        )
        self.FindImagesExist = channel.unary_unary(
            "/GeneralAPI/FindImagesExist",
            request_serializer=worker__api__pb2.ImagesHashes.SerializeToString,
            response_deserializer=worker__api__pb2.ImagesHashes.FromString,
        )
        self.AddExistingImagesToDataset = channel.unary_unary(
            "/GeneralAPI/AddExistingImagesToDataset",
            request_serializer=worker__api__pb2.ImagesToAdd.SerializeToString,
            response_deserializer=worker__api__pb2.ImageArray.FromString,
        )
        self.UploadAnnotations = channel.stream_unary(
            "/GeneralAPI/UploadAnnotations",
            request_serializer=worker__api__pb2.ChunkImage.SerializeToString,
            response_deserializer=worker__api__pb2.ImageArray.FromString,
        )
        self.UploadArchive = channel.stream_unary(
            "/GeneralAPI/UploadArchive",
            request_serializer=worker__api__pb2.Chunk.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )
        self.GetProjectStats = channel.unary_unary(
            "/GeneralAPI/GetProjectStats",
            request_serializer=worker__api__pb2.Id.SerializeToString,
            response_deserializer=worker__api__pb2.ProjectStats.FromString,
        )
        self.SetProjectFinished = channel.unary_unary(
            "/GeneralAPI/SetProjectFinished",
            request_serializer=worker__api__pb2.Id.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )
        self.Log = channel.unary_unary(
            "/GeneralAPI/Log",
            request_serializer=worker__api__pb2.LogLines.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )
        self.GetImportStructure = channel.unary_unary(
            "/GeneralAPI/GetImportStructure",
            request_serializer=worker__api__pb2.Id.SerializeToString,
            response_deserializer=worker__api__pb2.ListFiles.FromString,
        )
        self.GetImportFiles = channel.unary_stream(
            "/GeneralAPI/GetImportFiles",
            request_serializer=worker__api__pb2.ImportRequest.SerializeToString,
            response_deserializer=worker__api__pb2.ChunkFile.FromString,
        )
        self.GetNewTask = channel.unary_stream(
            "/GeneralAPI/GetNewTask",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.Task.FromString,
        )
        self.GetStopTask = channel.unary_stream(
            "/GeneralAPI/GetStopTask",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.Id.FromString,
        )
        self.AgentConnected = channel.unary_unary(
            "/GeneralAPI/AgentConnected",
            request_serializer=worker__api__pb2.AgentInfo.SerializeToString,
            response_deserializer=worker__api__pb2.ServerInfo.FromString,
        )
        self.AgentPing = channel.unary_unary(
            "/GeneralAPI/AgentPing",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )
        self.UploadModel = channel.stream_unary(
            "/GeneralAPI/UploadModel",
            request_serializer=worker__api__pb2.ChunkModel.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )
        self.DownloadModel = channel.unary_stream(
            "/GeneralAPI/DownloadModel",
            request_serializer=worker__api__pb2.ModelDescription.SerializeToString,
            response_deserializer=worker__api__pb2.Chunk.FromString,
        )
        self.GenerateNewModelId = channel.unary_unary(
            "/GeneralAPI/GenerateNewModelId",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.ModelDescription.FromString,
        )
        self.GetTelemetryTask = channel.unary_stream(
            "/GeneralAPI/GetTelemetryTask",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.Task.FromString,
        )
        self.UpdateTelemetry = channel.unary_unary(
            "/GeneralAPI/UpdateTelemetry",
            request_serializer=worker__api__pb2.AgentInfo.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )
        self.AddImages = channel.unary_unary(
            "/GeneralAPI/AddImages",
            request_serializer=worker__api__pb2.ImagesInfo.SerializeToString,
            response_deserializer=worker__api__pb2.ImageArray.FromString,
        )
        self.UploadImages = channel.stream_unary(
            "/GeneralAPI/UploadImages",
            request_serializer=worker__api__pb2.ChunkImage.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )
        self.GetUsedImageList = channel.unary_stream(
            "/GeneralAPI/GetUsedImageList",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.NodeObjectHash.FromString,
        )
        self.GetUsedModelList = channel.unary_stream(
            "/GeneralAPI/GetUsedModelList",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.NodeObjectHash.FromString,
        )
        self.GetGeneralEventsStream = channel.unary_stream(
            "/GeneralAPI/GetGeneralEventsStream",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.GeneralEvent.FromString,
        )
        self.GetGeneralEventData = channel.unary_stream(
            "/GeneralAPI/GetGeneralEventData",
            request_serializer=worker__api__pb2.Empty.SerializeToString,
            response_deserializer=worker__api__pb2.Chunk.FromString,
        )
        self.SendGeneralEventData = channel.stream_unary(
            "/GeneralAPI/SendGeneralEventData",
            request_serializer=worker__api__pb2.Chunk.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )
        self.AddMetaToProject = channel.unary_unary(
            "/GeneralAPI/AddMetaToProject",
            request_serializer=worker__api__pb2.Project.SerializeToString,
            response_deserializer=worker__api__pb2.Empty.FromString,
        )


class GeneralAPIServicer(object):
  """package supervisely;

  """

  def CreateProject(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def CreateDataset(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetProjectByName(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetDatasetByName(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetModelByName(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetProjectMeta(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetProjectDatasets(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetDatasetImages(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetImagesInfo(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DownloadImages(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DownloadAnnotations(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def FindImagesExist(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddExistingImagesToDataset(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UploadAnnotations(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UploadArchive(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetProjectStats(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SetProjectFinished(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def Log(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetImportStructure(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetImportFiles(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetNewTask(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetStopTask(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AgentConnected(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AgentPing(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UploadModel(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def DownloadModel(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GenerateNewModelId(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetTelemetryTask(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UpdateTelemetry(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddImages(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def UploadImages(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetUsedImageList(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetUsedModelList(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetGeneralEventsStream(self, request, context):
    """AgentRPC
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def GetGeneralEventData(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def SendGeneralEventData(self, request_iterator, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def AddMetaToProject(self, request, context):
    # missing associated documentation comment in .proto file
    pass
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_GeneralAPIServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'CreateProject': grpc.unary_unary_rpc_method_handler(
          servicer.CreateProject,
          request_deserializer=worker__api__pb2.Project.FromString,
          response_serializer=worker__api__pb2.Id.SerializeToString,
      ),
      'CreateDataset': grpc.unary_unary_rpc_method_handler(
          servicer.CreateDataset,
          request_deserializer=worker__api__pb2.ProjectDataset.FromString,
          response_serializer=worker__api__pb2.Id.SerializeToString,
      ),
      'GetProjectByName': grpc.unary_unary_rpc_method_handler(
          servicer.GetProjectByName,
          request_deserializer=worker__api__pb2.Project.FromString,
          response_serializer=worker__api__pb2.Project.SerializeToString,
      ),
      'GetDatasetByName': grpc.unary_unary_rpc_method_handler(
          servicer.GetDatasetByName,
          request_deserializer=worker__api__pb2.ProjectDataset.FromString,
          response_serializer=worker__api__pb2.Dataset.SerializeToString,
      ),
      'GetModelByName': grpc.unary_unary_rpc_method_handler(
          servicer.GetModelByName,
          request_deserializer=worker__api__pb2.ModelName.FromString,
          response_serializer=worker__api__pb2.ModelInfo.SerializeToString,
      ),
      'GetProjectMeta': grpc.unary_unary_rpc_method_handler(
          servicer.GetProjectMeta,
          request_deserializer=worker__api__pb2.Id.FromString,
          response_serializer=worker__api__pb2.Project.SerializeToString,
      ),
      'GetProjectDatasets': grpc.unary_unary_rpc_method_handler(
          servicer.GetProjectDatasets,
          request_deserializer=worker__api__pb2.Id.FromString,
          response_serializer=worker__api__pb2.DatasetArray.SerializeToString,
      ),
      'GetDatasetImages': grpc.unary_unary_rpc_method_handler(
          servicer.GetDatasetImages,
          request_deserializer=worker__api__pb2.Id.FromString,
          response_serializer=worker__api__pb2.ImageArray.SerializeToString,
      ),
      'GetImagesInfo': grpc.unary_unary_rpc_method_handler(
          servicer.GetImagesInfo,
          request_deserializer=worker__api__pb2.ImageArray.FromString,
          response_serializer=worker__api__pb2.ImagesInfo.SerializeToString,
      ),
      'DownloadImages': grpc.unary_stream_rpc_method_handler(
          servicer.DownloadImages,
          request_deserializer=worker__api__pb2.ImagesHashes.FromString,
          response_serializer=worker__api__pb2.ChunkImage.SerializeToString,
      ),
      'DownloadAnnotations': grpc.unary_stream_rpc_method_handler(
          servicer.DownloadAnnotations,
          request_deserializer=worker__api__pb2.ImageArray.FromString,
          response_serializer=worker__api__pb2.ChunkImage.SerializeToString,
      ),
      'FindImagesExist': grpc.unary_unary_rpc_method_handler(
          servicer.FindImagesExist,
          request_deserializer=worker__api__pb2.ImagesHashes.FromString,
          response_serializer=worker__api__pb2.ImagesHashes.SerializeToString,
      ),
      'AddExistingImagesToDataset': grpc.unary_unary_rpc_method_handler(
          servicer.AddExistingImagesToDataset,
          request_deserializer=worker__api__pb2.ImagesToAdd.FromString,
          response_serializer=worker__api__pb2.ImageArray.SerializeToString,
      ),
      'UploadAnnotations': grpc.stream_unary_rpc_method_handler(
          servicer.UploadAnnotations,
          request_deserializer=worker__api__pb2.ChunkImage.FromString,
          response_serializer=worker__api__pb2.ImageArray.SerializeToString,
      ),
      'UploadArchive': grpc.stream_unary_rpc_method_handler(
          servicer.UploadArchive,
          request_deserializer=worker__api__pb2.Chunk.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
      'GetProjectStats': grpc.unary_unary_rpc_method_handler(
          servicer.GetProjectStats,
          request_deserializer=worker__api__pb2.Id.FromString,
          response_serializer=worker__api__pb2.ProjectStats.SerializeToString,
      ),
      'SetProjectFinished': grpc.unary_unary_rpc_method_handler(
          servicer.SetProjectFinished,
          request_deserializer=worker__api__pb2.Id.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
      'Log': grpc.unary_unary_rpc_method_handler(
          servicer.Log,
          request_deserializer=worker__api__pb2.LogLines.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
      'GetImportStructure': grpc.unary_unary_rpc_method_handler(
          servicer.GetImportStructure,
          request_deserializer=worker__api__pb2.Id.FromString,
          response_serializer=worker__api__pb2.ListFiles.SerializeToString,
      ),
      'GetImportFiles': grpc.unary_stream_rpc_method_handler(
          servicer.GetImportFiles,
          request_deserializer=worker__api__pb2.ImportRequest.FromString,
          response_serializer=worker__api__pb2.ChunkFile.SerializeToString,
      ),
      'GetNewTask': grpc.unary_stream_rpc_method_handler(
          servicer.GetNewTask,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.Task.SerializeToString,
      ),
      'GetStopTask': grpc.unary_stream_rpc_method_handler(
          servicer.GetStopTask,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.Id.SerializeToString,
      ),
      'AgentConnected': grpc.unary_unary_rpc_method_handler(
          servicer.AgentConnected,
          request_deserializer=worker__api__pb2.AgentInfo.FromString,
          response_serializer=worker__api__pb2.ServerInfo.SerializeToString,
      ),
      'AgentPing': grpc.unary_unary_rpc_method_handler(
          servicer.AgentPing,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
      'UploadModel': grpc.stream_unary_rpc_method_handler(
          servicer.UploadModel,
          request_deserializer=worker__api__pb2.ChunkModel.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
      'DownloadModel': grpc.unary_stream_rpc_method_handler(
          servicer.DownloadModel,
          request_deserializer=worker__api__pb2.ModelDescription.FromString,
          response_serializer=worker__api__pb2.Chunk.SerializeToString,
      ),
      'GenerateNewModelId': grpc.unary_unary_rpc_method_handler(
          servicer.GenerateNewModelId,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.ModelDescription.SerializeToString,
      ),
      'GetTelemetryTask': grpc.unary_stream_rpc_method_handler(
          servicer.GetTelemetryTask,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.Task.SerializeToString,
      ),
      'UpdateTelemetry': grpc.unary_unary_rpc_method_handler(
          servicer.UpdateTelemetry,
          request_deserializer=worker__api__pb2.AgentInfo.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
      'AddImages': grpc.unary_unary_rpc_method_handler(
          servicer.AddImages,
          request_deserializer=worker__api__pb2.ImagesInfo.FromString,
          response_serializer=worker__api__pb2.ImageArray.SerializeToString,
      ),
      'UploadImages': grpc.stream_unary_rpc_method_handler(
          servicer.UploadImages,
          request_deserializer=worker__api__pb2.ChunkImage.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
      'GetUsedImageList': grpc.unary_stream_rpc_method_handler(
          servicer.GetUsedImageList,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.NodeObjectHash.SerializeToString,
      ),
      'GetUsedModelList': grpc.unary_stream_rpc_method_handler(
          servicer.GetUsedModelList,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.NodeObjectHash.SerializeToString,
      ),
      'GetGeneralEventsStream': grpc.unary_stream_rpc_method_handler(
          servicer.GetGeneralEventsStream,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.GeneralEvent.SerializeToString,
      ),
      'GetGeneralEventData': grpc.unary_stream_rpc_method_handler(
          servicer.GetGeneralEventData,
          request_deserializer=worker__api__pb2.Empty.FromString,
          response_serializer=worker__api__pb2.Chunk.SerializeToString,
      ),
      'SendGeneralEventData': grpc.stream_unary_rpc_method_handler(
          servicer.SendGeneralEventData,
          request_deserializer=worker__api__pb2.Chunk.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
      'AddMetaToProject': grpc.unary_unary_rpc_method_handler(
          servicer.AddMetaToProject,
          request_deserializer=worker__api__pb2.Project.FromString,
          response_serializer=worker__api__pb2.Empty.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'GeneralAPI', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# coding: utf-8

# docs
from __future__ import annotations

from copy import deepcopy
from typing import Dict, List, Optional, Union

import numpy as np

from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    ID,
    LABELER_LOGIN,
    UPDATED_AT,
)
from supervisely.geometry.geometry import Geometry
from supervisely.geometry.graph import EDGES, GraphNodes, Node, _maybe_transform_colors
from supervisely.geometry.point_location import PointLocation
from supervisely.imaging.color import _validate_color, hex2rgb, rgb2hex

VERTICES = "vertices"
CUBOID2D_VERTICES_NAMES = [
    "face1-topleft",
    "face1-topright",
    "face1-bottomright",
    "face1-bottomleft",
    "face2-topleft",
    "face2-topright",
    "face2-bottomright",
    "face2-bottomleft",
]
CUBOID2D_EDGES_MAPPING = [
    CUBOID2D_VERTICES_NAMES[:4],
    CUBOID2D_VERTICES_NAMES[4:],
    ["face1-topleft", "face2-topleft"],
    ["face1-topright", "face2-topright"],
    ["face1-bottomright", "face2-bottomright"],
    ["face1-bottomleft", "face2-bottomleft"],
]


class Cuboid2d(GraphNodes):
    """2D projection of cuboid as a graph of keypoints (eight corners). Immutable."""

    items_json_field = VERTICES

    @staticmethod
    def geometry_name():
        return "cuboid_2d"

    def __init__(
        self,
        nodes: Union[Dict[str, Dict], List],
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
        position: Optional[Dict] = None,
        rotation: Optional[Dict] = None,
        dimensions: Optional[Dict] = None,
        face: Optional[List[str]] = None,
    ):
        """
        Cuboid2d geometry for a single :class:`~supervisely.annotation.label.Label`. :class:`~supervisely.geometry.cuboid_2d.Cuboid2d` class object is immutable.

        :param nodes: Dict or List containing nodes of graph
        :type nodes: dict
        :param sly_id: Cuboid2d ID in Supervisely server.
        :type sly_id: int, optional
        :param class_id: ID of ObjClass to which Cuboid2d belongs.
        :type class_id: int, optional
        :param labeler_login: Login of the user who created :class:`~supervisely.geometry.cuboid_2d.Cuboid2d`.
        :type labeler_login: str, optional
        :param updated_at: Date and Time when Cuboid2d was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
        :type updated_at: str, optional
        :param created_at: Date and Time when Cuboid2d was created. Date Format is the same as in "updated_at" parameter.
        :type created_at: str, optional

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.graph import Node, Cuboid2d

                vertex_1 = Node(sly.PointLocation(5, 5))
                vertex_2 = Node(sly.PointLocation(100, 100))
                vertex_3 = Node(sly.PointLocation(200, 250))
                nodes = {0: vertex_1, 1: vertex_2, 2: vertex_3}
                figure = Cuboid2d(nodes)
        """
        super().__init__(
            nodes=nodes,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )
        self._position = position
        self._rotation = rotation
        self._dimensions = dimensions
        self._face = face

        if len(self._nodes) != 8:
            raise ValueError("Cuboid2d must have exactly 8 vertices")

    @property
    def vertices(self) -> Dict:
        """
        Copy of Cuboid2d vertices.

        :returns: Cuboid2d vertices
        :rtype: Optional[Dict]
        """
        return self.nodes

    @property
    def position(self) -> Optional[Dict]:
        """
        Copy of the position of the Cuboid2d.

        :returns: Position of the :class:`~supervisely.geometry.cuboid_2d.Cuboid2d`
        :rtype: Optional[Dict]
        """
        if isinstance(self._position, dict):
            return self._position.copy()

    @property
    def rotation(self) -> Optional[Dict]:
        """
        Copy of the rotation of the Cuboid2d.

        :returns: Rotation of the :class:`~supervisely.geometry.cuboid_2d.Cuboid2d`
        :rtype: Optional[Dict]
        """
        if isinstance(self._rotation, dict):
            return self._rotation.copy()

    @property
    def dimensions(self) -> Optional[Dict]:
        """
        Copy of the dimensions of the Cuboid2d.

        :returns: Dimensions of the :class:`~supervisely.geometry.cuboid_2d.Cuboid2d`
        :rtype: dict
        """
        if isinstance(self._dimensions, dict):
            return self._dimensions.copy()

    @property
    def face(self) -> Optional[List[str]]:
        """
        Copy of the face of the Cuboid2d.

        :returns: Face of the :class:`~supervisely.geometry.cuboid_2d.Cuboid2d`
        :rtype: Optional[List[str]]
        """
        if isinstance(self._face, list):
            return self._face.copy()

    @classmethod
    def from_json(cls, data: Dict[str, Dict]) -> Cuboid2d:
        """
        Convert a json dict to Cuboid2d. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Cuboid2d in json format as a dict.
        :type data: Dict[str, Dict]
        :returns: Cuboid2d from json.
        :rtype: :class:`~supervisely.geometry.cuboid_2d.Cuboid2d`

        :Usage Example:

            .. code-block:: python

                figure_json = {
                    "vertices": {
                        "0": {"loc": [5, 5]},
                        "1": {"loc": [100, 100]},
                        "2": {"loc": [250, 200]},
                        "position": {
                            "x": 0.0657651107620552,
                            "y": -0.05634319555373257,
                            "z": 0.7267282757573887
                        },
                        "rotation": { "x": 0, "y": 0, "z": 0 },
                        "dimensions": {
                            "x": 0.1425456564648202,
                            "y": 0.1,
                            "z": 0.36738880874660756
                        },
                        "face": [
                            "face2-topleft",
                            "face2-topright",
                            "face2-bottomright",
                            "face2-bottomleft"
                        ]
                    }
                }
                from supervisely.geometry.graph import Cuboid2d
                figure = Cuboid2d.from_json(figure_json)
        """
        nodes = {
            node_id: Node.from_json(node_json)
            for node_id, node_json in data[cls.items_json_field].items()
        }
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        position = data.get("position", None)
        rotation = data.get("rotation", None)
        dimensions = data.get("dimensions", None)
        face = data.get("face", None)
        return cls(
            nodes=nodes,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
            position=position,
            rotation=rotation,
            dimensions=dimensions,
            face=face,
        )

    def to_json(self) -> Dict[str, Dict]:
        """
        Convert the Cuboid2d to list. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :returns: Json format as a dict
        :rtype: Dict[str, Dict]

        :Usage Example:

            .. code-block:: python

                import supervisely as sly
                from supervisely.geometry.graph import Node, Cuboid2d

                vertex_1 = Node(sly.PointLocation(5, 5))
                vertex_2 = Node(sly.PointLocation(100, 100))
                vertex_3 = Node(sly.PointLocation(200, 250))
                nodes = {0: vertex_1, 1: vertex_2, 2: vertex_3}
                figure = Cuboid2d(nodes)

                figure_json = figure.to_json()
                print(figure_json)
                # Output: {
                #    "nodes": {
                #        "0": {
                #            "loc": [5, 5]
                #        },
                #        "1": {
                #            "loc": [100, 100]
                #        },
                #        "2": {
                #            "loc": [250, 200]
                #        }
                #    },
                #    "position": {
                #         "x": 0.0657651107620552,
                #         "y": -0.05634319555373257,
                #         "z": 0.7267282757573887
                #     },
                #     "rotation": { "x": 0, "y": 0, "z": 0 },
                #     "dimensions": {
                #         "x": 0.1425456564648202,
                #         "y": 0.1,
                #         "z": 0.36738880874660756
                #     },
                #     "face": [
                #         "face2-topleft",
                #         "face2-topright",
                #         "face2-bottomright",
                #         "face2-bottomleft"
                #     ],
                # }
        """
        res = {
            self.items_json_field: {
                node_id: node.to_json() for node_id, node in self._nodes.items()
            }
        }
        if self._position is not None:
            res["position"] = self._position
        if self._rotation is not None:
            res["rotation"] = self._rotation
        if self._dimensions is not None:
            res["dimensions"] = self._dimensions
        if self._face is not None:
            res["face"] = self._face

        self._add_creation_info(res)
        return res

    def validate(self, name: str, settings: Dict) -> None:
        """
        Checks the graph for correctness and compliance with the template
        """

        missing_nodes = set(settings[self.items_json_field].keys()) - set(self._nodes.keys())
        if len(missing_nodes) > 0:
            raise ValueError(f"Missing vertices in the Cuboid2d: {missing_nodes}.")
        if len(self._nodes) != 8:
            raise ValueError("Cuboid2d must have exactly 8 vertices")

        super().validate(name, settings)

    @staticmethod
    def _transform_config_colors(config, transform_fn):
        """
        Transform colors of edges and nodes in graph template
        :param config: dictionary(graph template)
        :param transform_fn: function to convert
        :returns: dictionary(graph template)
        """
        if config is None:
            return None

        result = deepcopy(config)
        _maybe_transform_colors(result.get(EDGES, []), transform_fn)
        _maybe_transform_colors(result[VERTICES].values(), transform_fn)
        return result

    @staticmethod
    def config_from_json(config: Dict) -> Dict:
        """
        Convert graph template from json format

        :param config: dictionary(graph template) in json format
        :type config: dict
        :returns: dictionary(graph template) in json format
        :rtype: dict
        """

        try:
            return Cuboid2d._transform_config_colors(config, hex2rgb)
        except Exception as e:
            raise RuntimeError(
                f"Failed to parse graph template from JSON format. "
                "Check out an example of a graph template in JSON format at: "
                "https://developer.supervisely.com/getting-started/python-sdk-tutorials/images/keypoints#click-to-see-the-example-of-template-in-json-format"
            )

    @staticmethod
    def config_to_json(config: Dict) -> Dict:
        """
        Convert graph template to json format

        :param config: dictionary(graph template)
        :type config: dict
        :returns: dictionary(graph template) in json format
        """
        return Cuboid2d._transform_config_colors(config, rgb2hex)

    @classmethod
    def allowed_transforms(cls):
        """
        Returns the allowed transforms for the Cuboid2d.
        """
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.rectangle import Rectangle

        return [AnyGeometry, Rectangle, GraphNodes]


class Cuboid2dTemplate(Cuboid2d, Geometry):
    """
    Geometry Config Template for a single :class:`~supervisely.geometry.cuboid_2d.Cuboid2d`. :class:`~supervisely.geometry.cuboid_2d.Cuboid2dTemplate` class object is immutable.
    """

    def __init__(self, color: List[int]):
        """
        :param color: RGB color for template vertices [r, g, b].
        :type color: List[int]
        """
        _validate_color(color)
        self._point_names = []
        self._config = self._create_template(color)

    def _create_template(self, color: List[int]) -> Cuboid2dTemplate:
        """
        Returns a template for a single :class:`~supervisely.geometry.cuboid_2d.Cuboid2d`.
        """
        config = {VERTICES: {}, EDGES: []}

        x = y = w = h = s = 1  # sample values only for config creation
        base_vertices = [(x, y), (x + w, y), (x + w, y + h), (x, y + h)]
        shifted_vertices = [(vx + s, vy + s) for vx, vy in base_vertices]
        verices_coords = base_vertices + shifted_vertices

        for label, coords in zip(CUBOID2D_VERTICES_NAMES, verices_coords):
            col, row = coords
            self._point_names.append(label)
            config[VERTICES][label] = {"label": label, "loc": [row, col], "color": color}

        for edges in CUBOID2D_EDGES_MAPPING:
            if len(edges) == 2:
                config[EDGES].append({"src": edges[0], "dst": edges[1], "color": color})
            else:
                for i in range(len(edges)):
                    config[EDGES].append(
                        {"src": edges[i], "dst": edges[(i + 1) % len(edges)], "color": color}
                    )

        return config

    def get_nodes(self):
        self._nodes = {}
        for node in self._config[self.items_json_field]:
            loc = self._config[self.items_json_field][node]["loc"]
            self._nodes[node] = Node(PointLocation(loc[1], loc[0]), label=node)

    def draw(self, image: np.ndarray, thickness=7):
        self.get_nodes()
        self._draw_bool_compatible(
            self._draw_impl,
            bitmap=image,
            color=[0, 255, 0],
            thickness=thickness,
            config=self._config,
        )

    def to_json(self):
        return self.config_to_json(self._config)

    @property
    def config(self):
        return self._config

    @property
    def point_names(self):
        """
        Return point names in order in which they were added
        """
        return self._point_names



from __future__ import annotations

import os
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

from tqdm import tqdm

from supervisely._utils import batched, get_or_create_event_loop, is_production
from supervisely.annotation.annotation import Annotation
from supervisely.annotation.obj_class import ObjClass
from supervisely.annotation.tag_meta import TagValueType
from supervisely.api.api import Api
from supervisely.geometry.graph import GraphNodes
from supervisely.io.env import team_id
from supervisely.io.fs import (
    get_file_ext,
    get_file_name_with_ext,
    is_archive,
    remove_dir,
    silent_remove,
    unpack_archive,
)
from supervisely.project.project_meta import ProjectMeta
from supervisely.project.project_settings import LabelingInterface
from supervisely.sly_logger import logger
from supervisely.task.progress import Progress


class AvailableImageConverters:
    """Names of supported image dataset converters."""

    SLY = "supervisely"
    COCO = "coco"
    FAST_COCO = "coco (fast)"
    YOLO = "yolo"
    PASCAL_VOC = "pascal_voc"
    CSV = "csv"
    MULTISPECTRAL = "multispectral"
    MASKS = "images_with_masks"
    MEDICAL2D = "medical_2d"
    MULTI_VIEW = "multi_view"
    PDF = "pdf"
    CITYSCAPES = "cityscapes"
    LABEL_ME = "label_me"
    LABEL_STUDIO = "label_studio"
    HIGH_COLOR_DEPTH = "high_color_depth"


class AvailableVideoConverters:
    """Names of supported video dataset converters."""

    SLY = "supervisely"
    MOT = "mot"
    DAVIS = "davis"
    MULTI_VIEW = "multi_view"


class AvailablePointcloudConverters:
    """Names of supported point cloud dataset converters."""

    SLY = "supervisely"
    LAS = "las/laz"
    PLY = "ply"
    BAG = "rosbag"
    LYFT = "lyft"
    NUSCENES = "nuscenes"
    KITTI3D = "kitti3d"


class AvailablePointcloudEpisodesConverters:
    """Names of supported point cloud episode dataset converters."""

    SLY = "supervisely"
    BAG = "rosbag"
    LYFT = "lyft"
    KITTI360 = "kitti360"


class AvailableVolumeConverters:
    """Names of supported 3D volume dataset converters."""

    SLY = "supervisely"
    DICOM = "dicom"
    NII = "nii"


class BaseConverter:
    """
    Base class for dataset format converters into Supervisely projects.

    Detects input format, validates annotations, builds :class:`~supervisely.project.project_meta.ProjectMeta`
    and provides methods to upload converted data.
    """
    unsupported_exts = [".gif", ".html", ".htm"]

    class BaseItem:
        """Represents a single convertible item (data path + raw annotation data + optional shape/custom_data)."""

        def __init__(
            self,
            item_path: str,
            ann_data: Union[str, dict] = None,
            shape: Union[Tuple, List] = None,
            custom_data: Optional[dict] = None,
        ):
            """:param item_path: Path to the item file.
            :type item_path: str
            :param ann_data: Raw annotation (path, dict, or JSON string).
            :type ann_data: Union[str, dict], optional
            :param shape: Image/item shape (height, width).
            :type shape: Union[Tuple, List], optional
            :param custom_data: Extra per-item data.
            :type custom_data: dict, optional
            """
            self._path: str = item_path
            self._name: str = None
            self._ann_data: Union[str, dict, list] = ann_data
            self._shape: Union[Tuple, List] = shape
            self._custom_data: dict = custom_data or {}

        @property
        def name(self) -> str:
            if self._name is not None:
                return self._name
            return get_file_name_with_ext(self._path)

        @name.setter
        def name(self, name: str) -> None:
            self._name = name

        @property
        def path(self) -> str:
            return self._path

        @path.setter
        def path(self, path: str) -> None:
            self._path = path

        @property
        def ann_data(self) -> Union[str, dict]:
            return self._ann_data

        @ann_data.setter
        def ann_data(self, ann_data: Union[str, dict, list]) -> None:
            self._ann_data = ann_data

        @property
        def shape(self) -> Union[Tuple, List]:
            return self._shape

        @property
        def custom_data(self) -> dict:
            return self._custom_data

        @custom_data.setter
        def custom_data(self, custom_data: dict) -> None:
            self._custom_data = custom_data

        def set_shape(self, shape) -> None:
            self._shape = shape

        def update_custom_data(self, custom_data: dict) -> None:
            self.custom_data.update(custom_data)

        def update(
            self,
            item_path: str = None,
            ann_data: Union[str, dict] = None,
            shape: Union[Tuple, List] = None,
            custom_data: dict = {},
        ) -> None:
            if item_path is not None:
                self.path = item_path
            if ann_data is not None:
                self.ann_data = ann_data
            if shape is not None:
                self.set_shape(shape)
            if custom_data:
                self.update_custom_data(custom_data)

        def create_empty_annotation(self) -> Annotation:
            raise NotImplementedError()

    def __init__(
        self,
        input_data: str,
        labeling_interface: Optional[Union[LabelingInterface, str]] = LabelingInterface.DEFAULT,
        upload_as_links: bool = False,
        remote_files_map: Optional[Dict[str, str]] = None,
    ):
        """:param input_data: Path to input directory or archive.
        :type input_data: str
        :param labeling_interface: Labeling interface preset.
        :type labeling_interface: LabelingInterface, optional
        :param upload_as_links: If True, upload as Team Files links.
        :type upload_as_links: bool
        :param remote_files_map: Map of local paths to remote paths for link upload.
        :type remote_files_map: Dict[str, str], optional

        :raises ValueError: If labeling_interface is invalid.
        """
        self._input_data: str = input_data
        self._items: List[BaseConverter.BaseItem] = []
        self._meta: ProjectMeta = None
        self._labeling_interface = labeling_interface or LabelingInterface.DEFAULT

        # import as links settings
        self._upload_as_links: bool = upload_as_links
        self._remote_files_map: Optional[Dict[str, str]] = remote_files_map
        self._supports_links = False  # if converter supports uploading by links
        self._force_shape_for_links = False
        self._api = Api.from_env() if self._upload_as_links else None
        self._team_id = team_id() if self._upload_as_links else None
        self._converter = None

        if self._labeling_interface not in LabelingInterface.values():
            raise ValueError(
                f"Invalid labeling interface value: {labeling_interface}. "
                f"The available values: {LabelingInterface.values()}"
            )

    @property
    def format(self) -> str:
        return self.__str__()

    @property
    def items_count(self) -> int:
        return len(self._items)

    @property
    def ann_ext(self) -> str:
        raise NotImplementedError()

    @property
    def key_file_ext(self) -> str:
        raise NotImplementedError()

    @property
    def upload_as_links(self) -> bool:
        return self._upload_as_links

    @property
    def remote_files_map(self) -> Dict[str, str]:
        return self._remote_files_map

    @property
    def supports_links(self) -> bool:
        return self._supports_links

    def validate_labeling_interface(self) -> bool:
        return self._labeling_interface == LabelingInterface.DEFAULT

    def validate_ann_file(self, ann_path) -> bool:
        raise NotImplementedError()

    def validate_key_file(self) -> bool:
        raise NotImplementedError()

    def detect_format(self) -> BaseConverter:
        self._converter = self._detect_format()
        return self._converter

    def validate_format(self) -> bool:
        """
        Validate format of the input data meets the requirements of the converter. Should be implemented in the subclass.
        Additionally, this method must do the following steps:
            1. creates project meta (if key file file exists) and save it to self._meta
            2. creates items, count detected annotations and save them to self._items
            3. validates annotation files (and genereate meta if key file is missing)

        :returns: True if format is valid, False otherwise.
        """
        raise NotImplementedError()

    def get_meta(self) -> ProjectMeta:
        return self._meta

    def get_items(self) -> List[BaseItem]:
        return self._items

    def to_supervisely(self, item: BaseItem, meta: ProjectMeta, *args) -> Annotation:
        """Convert to Supervisely format."""
        return item.create_empty_annotation()

    def upload_dataset(self, api: Api, dataset_id: int, batch_size: int = 50) -> None:
        """Upload converted data to Supervisely"""
        raise NotImplementedError()

    def _detect_format(self):
        found_formats = []
        all_converters = self.__class__.__subclasses__()

        progress, progress_cb = self.get_progress(1, "Detecting annotation format")
        for converter in all_converters:
            if converter.__name__ == "BaseConverter":
                continue
            converter = converter(
                self._input_data,
                self._labeling_interface,
                self._upload_as_links,
                self._remote_files_map,
            )

            if not converter.validate_labeling_interface():
                continue

            if self.upload_as_links and not converter.supports_links:
                continue

            if converter.validate_format():
                logger.info(f"Detected format: {str(converter)}")
                found_formats.append(converter)
                if len(found_formats) > 1:
                    raise RuntimeError(
                        f"Multiple formats detected: {[str(f) for f in found_formats]}. "
                        "Mixed formats are not supported yet."
                    )

        progress_cb(1)

        if len(found_formats) == 0:
            self._items, only_modality_items, unsupported_exts = (
                self._collect_items_if_format_not_detected()
            )

            if self.items_count == 0:
                if unsupported_exts:
                    raise RuntimeError(
                        f"Not found any {self.modality} to upload. "  # pylint: disable=no-member
                        f"Unsupported file extensions detected: {unsupported_exts}. "
                        f"Convert your data to one of the supported formats: {self.allowed_exts}"  # pylint: disable=no-member
                    )
                raise RuntimeError(
                    "Please refer to the app overview and documentation for annotation formats, "
                    "and ensure that your data contains valid information"
                )
            if not only_modality_items:
                logger.warning(
                    "Annotations not found. "  # pylint: disable=no-member
                    f"Uploading {self.modality} without annotations. "
                    "If you need assistance to upload data with annotations, please contact our support team."
                )
            return self

        if len(found_formats) == 1:
            return found_formats[0]

    def _collect_items_if_format_not_detected(self):
        only_modality_items = True
        unsupported_exts = set()
        items = []
        for root, _, files in os.walk(self._input_data):
            for file in files:
                full_path = os.path.join(root, file)
                ext = get_file_ext(full_path)
                if ext.lower() in self.allowed_exts:  # pylint: disable=no-member
                    items.append(self.Item(full_path))  # pylint: disable=no-member
                    continue
                only_modality_items = False
                if ext.lower() in self.unsupported_exts:
                    unsupported_exts.add(ext)

        return items, only_modality_items, unsupported_exts

    def merge_metas_with_conflicts(
        self, api: Api, dataset_id: int
    ) -> Tuple[ProjectMeta, dict, dict]:

        # get meta1 from project and meta2 from converter
        dataset = api.dataset.get_info_by_id(dataset_id)
        if dataset is None:
            raise RuntimeError(
                f"Dataset ID:{dataset_id} not found. "
                "Please check if the dataset exists and try again."
            )
        meta1_json = api.project.get_meta(dataset.project_id, with_settings=True)
        meta1 = ProjectMeta.from_json(meta1_json)
        meta2 = self._meta

        if meta2 is None:
            return meta1, {}, {}

        # merge classes and tags from meta1 (unchanged) and meta2 (renamed if conflict)
        renamed_classes = {}
        renamed_tags = {}
        for new_cls in meta2.obj_classes:
            i = 1
            new_name = new_cls.name
            matched = False
            def _is_matched(old_cls: ObjClass, new_cls: ObjClass) -> bool:
                if old_cls.geometry_type == new_cls.geometry_type:
                    if old_cls.geometry_type == GraphNodes:
                        old_nodes = old_cls.geometry_config["nodes"]
                        new_nodes = new_cls.geometry_config["nodes"]
                        return old_nodes.keys() == new_nodes.keys()
                    return True
                return False

            while meta1.obj_classes.get(new_name) is not None:
                if _is_matched(meta1.get_obj_class(new_name), new_cls):
                    matched = True
                    break
                new_name = f"{new_cls.name}_{i}"
                i += 1
            if new_name != new_cls.name:
                logger.warning(f"Class {new_cls.name} renamed to {new_name}")
                renamed_classes[new_cls.name] = new_name
            if not matched:
                new_cls = new_cls.clone(name=new_name)
                meta1 = meta1.add_obj_class(new_cls)

        for new_tag in meta2.tag_metas:
            i = 1
            new_name = new_tag.name
            matched = False
            while meta1.tag_metas.get(new_name) is not None:
                if meta1.tag_metas.get(new_name).value_type == new_tag.value_type:
                    if new_tag.value_type != TagValueType.ONEOF_STRING:
                        matched = True
                        break
                    if meta1.tag_metas.get(new_name).possible_values == new_tag.possible_values:
                        matched = True
                        break
                new_name = f"{new_tag.name}_{i}"
                i += 1
            if new_name != new_tag.name:
                logger.warning(f"Tag {new_tag.name} renamed to {new_name}")
                renamed_tags[new_tag.name] = new_name
            if not matched:
                new_tag = new_tag.clone(name=new_name)
                meta1 = meta1.add_tag_meta(new_tag)

        meta1 = self._update_labeling_interface(meta1, meta2, renamed_tags)

        # update project meta
        meta1 = api.project.update_meta(dataset.project_id, meta1)

        return meta1, renamed_classes, renamed_tags

    def get_progress(
        self,
        total: int,
        message: str = "Processing items...",
        is_size: bool = False,
    ) -> tuple:
        if is_production():
            progress = Progress(message, total, is_size=is_size)
            progress_cb = progress.iters_done_report
        else:
            progress = tqdm(
                total=total, desc=message, unit="B" if is_size else "it", unit_scale=is_size
            )
            progress_cb = progress.update
        return progress, progress_cb

    def _update_labeling_interface(
        self,
        meta1: ProjectMeta,
        meta2: ProjectMeta,
        renamed_tags: Dict[str, str] = None,
    ) -> ProjectMeta:
        """
        Update project meta with labeling interface from the converter meta.
        Only update if the existing labeling interface is the default value.
        In other cases, the existing labeling interface is preserved.
        """
        existing = meta1.project_settings.labeling_interface
        new = meta2.project_settings.labeling_interface
        if existing == new or new == LabelingInterface.DEFAULT:
            return meta1

        group_tag_name = meta2.project_settings.multiview_tag_name
        if group_tag_name:
            group_tag_name = renamed_tags.get(group_tag_name, group_tag_name)
            new_settings = meta2.project_settings.clone(multiview_tag_name=group_tag_name)
        else:
            new_settings = meta2.project_settings

        return meta1.clone(project_settings=new_settings)

    def _download_remote_ann_files(self) -> None:
        """
        Download all annotation files from Cloud Storage to the local storage.
        Needed to detect annotation format if "upload_as_links" is enabled.
        """
        if not self.upload_as_links:
            return

        ann_archives = {l: r for l, r in self._remote_files_map.items() if is_archive(l)}

        anns_to_download = {
            l: r for l, r in self._remote_files_map.items() if get_file_ext(l) == self.ann_ext
        }
        if not anns_to_download and not ann_archives:
            return

        import asyncio

        for files_type, files in {
            "annotations": anns_to_download,
            "archives": ann_archives,
        }.items():
            if not files:
                continue

            is_archive_type = files_type == "archives"

            file_size = None
            if is_archive_type:
                logger.info(f"Remote archives detected.")
                file_size = sum(
                    self._api.storage.get_info_by_path(self._team_id, remote_path).sizeb
                    for remote_path in files.values()
                )

            loop = get_or_create_event_loop()
            _, progress_cb = self.get_progress(
                len(files) if not is_archive_type else file_size,
                f"Downloading {files_type} from remote storage",
                is_size=is_archive_type,
            )

            for local_path in files.keys():
                silent_remove(local_path)

            logger.info(f"Downloading {files_type} from remote storage...")
            download_coro = self._api.storage.download_bulk_async(
                team_id=self._team_id,
                remote_paths=list(files.values()),
                local_save_paths=list(files.keys()),
                progress_cb=progress_cb,
                progress_cb_type="number" if not is_archive_type else "size",
            )

            if loop.is_running():
                future = asyncio.run_coroutine_threadsafe(download_coro, loop=loop)
                future.result()
            else:
                loop.run_until_complete(download_coro)
            logger.info("Possible annotations downloaded successfully.")

            if is_archive_type:
                for local_path in files.keys():
                    parent_dir = Path(local_path).parent
                    if parent_dir.name == "ann":
                        target_dir = parent_dir
                    else:
                        target_dir = parent_dir / "ann"
                        target_dir.mkdir(parents=True, exist_ok=True)

                    unpack_archive(local_path, str(target_dir))
                    silent_remove(local_path)

                    dirs = [d for d in target_dir.iterdir() if d.is_dir()]
                    files = [f for f in target_dir.iterdir() if f.is_file()]
                    if len(dirs) == 1 and len(files) == 0:
                        for file in dirs[0].iterdir():
                            file.rename(target_dir / file.name)
                        remove_dir(str(dirs[0]))

                    logger.info(f"Archive {local_path} unpacked successfully to {str(target_dir)}")



import os
import subprocess
from typing import Dict, Optional, Tuple, Union

import cv2
import magic

from supervisely import (
    Api,
    KeyIdMap,
    Progress,
    VideoAnnotation,
    batched,
    generate_free_name,
    is_development,
    logger,
)
from supervisely.convert.base_converter import BaseConverter
from supervisely.io.fs import (
    get_file_ext,
    get_file_name,
    get_file_name_with_ext,
    get_file_size,
    silent_remove,
)
from supervisely.project.project_settings import LabelingInterface
from supervisely.video.video import ALLOWED_VIDEO_EXTENSIONS, get_info


class VideoConverter(BaseConverter):
    """Base converter for video projects (uploads videos and optional video annotations to Supervisely)."""

    allowed_exts = ALLOWED_VIDEO_EXTENSIONS + [".mpg"]
    base_video_extension = ".mp4"
    modality = "videos"

    class Item(BaseConverter.BaseItem):
        """Video item with cached shape/frame count and optional annotation/meta for upload."""

        def __init__(
            self,
            item_path,
            ann_data=None,
            shape=None,
            custom_data=None,
            frame_count=None,
            metadata=None,
        ):
            """
            :param item_path: Path to video file.
            :type item_path: str
            :param ann_data: Annotation path or data.
            :type ann_data: str, optional
            :param shape: (height, width). Inferred from video if None.
            :type shape: tuple, optional
            :param custom_data: Extra data.
            :type custom_data: dict, optional
            :param frame_count: Frame count. Inferred if None.
            :type frame_count: int, optional
            :param metadata: Optional metadata path.
            :type metadata: str, optional
            """
            self._path = item_path
            self._name: str = None
            self._ann_data = ann_data
            self._metadata = metadata
            self._type = "video"
            if shape is None:
                vcap = cv2.VideoCapture(item_path)
                if vcap.isOpened():
                    width = int(vcap.get(cv2.CAP_PROP_FRAME_WIDTH))
                    height = int(vcap.get(cv2.CAP_PROP_FRAME_HEIGHT))
                    self._shape = (height, width)
                    self._frame_count = int(vcap.get(cv2.CAP_PROP_FRAME_COUNT))
            else:
                self._shape = shape
                self._frame_count = frame_count
            self._custom_data = custom_data if custom_data is not None else {}

        @property
        def shape(self) -> Tuple[int, int]:
            return self._shape

        @shape.setter
        def shape(self, shape: Optional[Tuple[int, int]] = None):
            self._shape = shape if shape is not None else [None, None]

        @property
        def frame_count(self) -> int:
            return self._frame_count

        @frame_count.setter
        def frame_count(self, frame_count: int):
            self._frame_count = frame_count

        @property
        def name(self) -> str:
            if self._name is not None:
                return self._name
            return get_file_name_with_ext(self._path)

        @name.setter
        def name(self, name: str):
            self._name = name

        @property
        def metadata(self) -> Optional[str]:
            return self._metadata

        @metadata.setter
        def metadata(self, metadata: Optional[str]):
            self._metadata = metadata

        def create_empty_annotation(self) -> VideoAnnotation:
            return VideoAnnotation(self._shape, self._frame_count)

    def __init__(
        self,
        input_data: str,
        labeling_interface: Optional[Union[LabelingInterface, str]],
        upload_as_links: bool,
        remote_files_map: Optional[Dict[str, str]] = None,
    ):
        """See :class:`~supervisely.convert.base_converter.BaseConverter` for params."""
        super().__init__(input_data, labeling_interface, upload_as_links, remote_files_map)
        self._key_id_map: KeyIdMap = None

    @property
    def format(self):
        return self._converter.format

    @property
    def ann_ext(self):
        return None

    @property
    def key_file_ext(self):
        return None

    @staticmethod
    def validate_ann_file(ann_path, meta=None):
        return False

    def upload_dataset(
        self,
        api: Api,
        dataset_id: int,
        batch_size: int = 10,
        log_progress=True,
    ):
        """Upload converted data to Supervisely"""
        meta, renamed_classes, renamed_tags = self.merge_metas_with_conflicts(api, dataset_id)
        videos_in_dataset = api.video.get_list(dataset_id, force_metadata_for_links=False)
        existing_names = {video_info.name for video_info in videos_in_dataset}

        # check video codecs, mimetypes and convert if needed
        convert_progress, convert_progress_cb = self.get_progress(
            self.items_count, "Preparing videos..."
        )
        for item in self._items:
            item_name, item_path = self.convert_to_mp4_if_needed(item.path)
            item.name = item_name
            item.path = item_path
            convert_progress_cb(1)
        if is_development():
            convert_progress.close()

        has_large_files = False
        size_progress_cb = None
        progress_cb, progress, ann_progress, ann_progress_cb = None, None, None, None
        if log_progress:
            progress, progress_cb = self.get_progress(self.items_count, "Uploading videos...")
            if not self.upload_as_links:
                file_sizes = [get_file_size(item.path) for item in self._items]
                has_large_files = any(
                    [self._check_video_file_size(file_size) for file_size in file_sizes]
                )
                if has_large_files:
                    upload_progress = []
                    size_progress_cb = self._get_video_upload_progress(upload_progress)
        batch_size = 1 if has_large_files and not self.upload_as_links else batch_size

        for batch in batched(self._items, batch_size=batch_size):
            item_names = []
            item_paths = []
            item_metas = []
            anns = []
            figures_cnt = 0
            for item in batch:
                item.name = generate_free_name(
                    existing_names, item.name, with_ext=True, extend_used_names=True
                )
                item_paths.append(item.path)
                item_names.append(item.name)

                if isinstance(item.metadata, str):  # path to file
                    from supervisely.io.json import load_json_file

                    item_metas.append(load_json_file(item.metadata))
                elif isinstance(item.metadata, dict):
                    item_metas.append(item.metadata)
                else:
                    item_metas.append({})

                ann = None
                if not self.upload_as_links or self.supports_links:
                    ann = self.to_supervisely(item, meta, renamed_classes, renamed_tags)
                    if ann is not None:
                        figures_cnt += len(ann.figures)
                anns.append(ann)

            if self.upload_as_links:
                vid_infos = api.video.upload_links(
                    dataset_id,
                    item_paths,
                    item_names,
                    metas=item_metas,
                    skip_download=True,
                    progress_cb=progress_cb if log_progress else None,
                    force_metadata_for_links=False,
                )
            else:
                vid_infos = api.video.upload_paths(
                    dataset_id,
                    item_names,
                    item_paths,
                    progress_cb=progress_cb if log_progress else None,
                    item_progress=(size_progress_cb if log_progress and has_large_files else None),
                    metas=item_metas,
                )
            vid_ids = [vid_info.id for vid_info in vid_infos]

            if log_progress and has_large_files and figures_cnt > 0:
                ann_progress, ann_progress_cb = self.get_progress(
                    figures_cnt, "Uploading annotations..."
                )

            for vid, ann, item, info in zip(vid_ids, anns, batch, vid_infos):
                if ann is None:
                    ann = VideoAnnotation((info.frame_height, info.frame_width), info.frames_count)
                api.video.annotation.append(vid, ann, progress_cb=ann_progress_cb)

        if log_progress and is_development():
            if progress is not None:
                progress.close()
            if ann_progress is not None:
                ann_progress.close()
        logger.info(f"Dataset ID:{dataset_id} has been successfully uploaded.")

    def convert_to_mp4_if_needed(self, video_path):
        video_name = get_file_name_with_ext(video_path)
        if self.upload_as_links:
            video_path = self.remote_files_map.get(video_path)
            return video_name, video_path
        video_ext = get_file_ext(video_path)
        if video_ext.lower() != video_ext:
            # rename video file to make extension lowercase
            new_video_path = os.path.splitext(video_path)[0] + video_ext.lower()
            os.rename(video_path, new_video_path)
            video_path = new_video_path

        # convert
        output_video_name = f"{get_file_name(video_name)}{self.base_video_extension}"
        output_video_path = os.path.splitext(video_path)[0] + "_h264" + self.base_video_extension

        # read video meta_data
        try:
            vid_meta = get_info(video_path)
            need_video_transc, need_audio_transc = self._check_codecs(vid_meta)
        except:
            need_video_transc, need_audio_transc = True, True

        if not need_video_transc:
            # check if video is already in mp4 format and mime type is `video/mp4`
            if video_path.lower().endswith(".mp4"):
                mime = magic.Magic(mime=True)
                mime_type = mime.from_file(video_path)
                if mime_type == "video/mp4":
                    logger.debug(
                        f'Video "{video_name}" is already in mp4 format, conversion is not required.'
                    )
                    return output_video_name, video_path
                else:
                    need_video_transc = True
            else:
                need_video_transc = True

        # convert videos
        self._convert(
            input_path=video_path,
            output_path=output_video_path,
            need_video_transc=need_video_transc,
            need_audio_transc=need_audio_transc,
        )
        if os.path.exists(output_video_path):
            logger.info(f"Video {video_name} has been converted.")
            silent_remove(video_path)

        return output_video_name, output_video_path

    def _check_codecs(self, video_meta):
        need_video_transc, need_audio_transc = False, False
        for stream in video_meta["streams"]:
            codec_type = stream["codecType"]
            if codec_type not in ["video", "audio"]:
                continue
            codec_name = stream["codecName"]
            if codec_type == "video" and codec_name not in ["h264", "h265", "hevc", "av1"]:
                logger.info(f"Video codec is not h264/h265/hevc/av1, transcoding is required: {codec_name}")
                need_video_transc = True
            elif codec_type == "audio" and codec_name != "aac":
                logger.info(f"Audio codec is not aac, transcoding is required: {codec_name}")
                need_audio_transc = True
        return need_video_transc, need_audio_transc

    def _convert(self, input_path, output_path, need_video_transc, need_audio_transc):
        video_codec = "libx264" if need_video_transc else "copy"
        audio_codec = "aac" if need_audio_transc else "copy"
        logger.info("Converting video...")
        subprocess.call(
            [
                "ffmpeg",
                "-y",
                "-i",
                f"{input_path}",
                "-c:v",
                f"{video_codec}",
                "-c:a",
                f"{audio_codec}",
                f"{output_path}",
            ]
        )

    def _check_video_file_size(self, file_size):
        return file_size > 20 * 1024 * 1024  # 20 MB

    def _get_video_upload_progress(self, upload_progress):
        upload_progress = []

        def _print_progress(monitor, upload_progress):
            if len(upload_progress) == 0:
                upload_progress.append(
                    Progress(
                        message="Upload videos...",
                        ext_logger=logger,
                        is_size=True,
                    )
                )
            upload_progress[0].set_current_value(monitor)

        return lambda m: _print_progress(m, upload_progress)



# Video
from supervisely.convert.video.mot.mot_converter import MOTConverter
from supervisely.convert.video.sly.sly_video_converter import SLYVideoConverter
from supervisely.convert.video.multi_view.multi_view import MultiViewVideoConverter



from supervisely.app.widgets_context import JinjaWidgets

from supervisely.annotation.annotation import Annotation, ProjectMeta
from supervisely.app import DataJson
from supervisely.app.widgets import Widget


class ImageAnnotationPreview(Widget):
    """Displays an image with overlaid annotation (labels, masks) and optional zoom; driven by set()."""

    def __init__(
        self,
        annotations_opacity: float = 0.5,
        enable_zoom: bool = False,
        line_width: int = 1,
        widget_id: str = None,
    ):
        """
        :param annotations_opacity: Opacity of annotation overlays (0-1).
        :type annotations_opacity: float
        :param enable_zoom: If True, enable zoom.
        :type enable_zoom: bool
        :param line_width: Line width for shapes.
        :type line_width: int
        :param widget_id: Unique widget identifier.
        :type widget_id: str, optional
        """
        self._image_url = None
        self._annotation = None
        self._project_meta = None
        self._opacity = annotations_opacity
        self._enable_zoom = enable_zoom
        self._line_width = line_width

        super().__init__(widget_id=widget_id, file_path=__file__)
        script_path = "./sly/css/app/widgets/image_annotation_preview/script.js"
        JinjaWidgets().context["__widget_scripts__"][self.__class__.__name__] = script_path

    def get_json_data(self):
        return {
            "imageUrl": self._image_url,
            "projectMeta": self._project_meta.to_json() if self._project_meta else None,
            "annotation": {"annotation": self._annotation.to_json()} if self._annotation else None,
        }

    def get_json_state(self):
        return {
            "options": {
                "enableZoom": self._enable_zoom,
                "opacity": self._opacity,
                "lineWidth": self._line_width,
                "fitOnResize": True,
            }
        }

    def set(
        self,
        image_url,
        ann: Annotation = None,
        project_meta: ProjectMeta = None,
    ):
        self.clean_up()
        self._image_url = image_url
        self._annotation = ann
        self._project_meta = project_meta
        self.update_data()
        DataJson().send_changes()

    def clean_up(self):
        self._image_shape = None
        self._image_url = None
        self._annotation = None
        self._project_meta = None
        self.update_data()
        DataJson().send_changes()

    def is_empty(self):
        return self._image_url is None



# coding: utf-8

# docs
from supervisely.pointcloud_annotation.pointcloud_episode_tag_collection import (
    PointcloudEpisodeTagCollection,
)
from supervisely.pointcloud_annotation.pointcloud_object import PointcloudObject


class PointcloudEpisodeObject(PointcloudObject):
    """Object tracked across point cloud episode frames (obj_class + tags). Uses VideoObject constructor. Immutable."""

    tag_collection_type = PointcloudEpisodeTagCollection
